#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

/* struct to hold the lines/columns to/not to print in a for(;;) like
 * structure */

typedef struct fields_ {
  int *from,*to,*inc;
  int num;		/* number of fields allocated */
} fields;

typedef struct cellstr_ {
  int type;		/* BP_Mask means absolute columns */
  long from, to, inc;	/* object number */
  fields *f;		/* fields of object */
} cellstr;

#define BT_EOL		0
#define BT_Wall		1
#define BT_Ball		2
#define BT_Cylinder	3
#define BT_Disk		4
#define BT_BallCC	5
#define BT_Force	6
#define BT_STBall	7
#define BT_DTEdge	8
#define BT_DGBall	9
#define BT_Ellipsoid	10
#define BT_Arrow	12
#define BT_Cone		13
#define BT_Triangle	14
#define BT_Triangle2	15

/* body parameter (type 4) */
#define BP_Mask		31
#define BP_Size		0
#define BP_Pos		32
#define BP_Orient	64
#define BP_Coord	128
#define BP_Velo		256
#define BP_Force	512
#define BP_AngVelo	1024
#define BP_Torque	2048
#define BP_TrVelo	4096
#define BP_TrForce	8192
#define BP_TrAngVelo	16384
#define BP_Handle	32768
#define BP_LifeTime	65536
#define BP_MatPar	131072
#define BP_TrCoord	262144
#define BP_UpMask	1048544	/* one more bit is needed!!! */
#define BP_N 15	/* maximum parameter an object can have update! XXX */

int par_typ[BP_N] = { BP_Size, BP_Pos, BP_Orient, BP_Coord, BP_Velo, BP_Force, 
		BP_AngVelo, BP_Torque, BP_TrVelo, BP_TrForce, BP_TrAngVelo,
		BP_Handle, BP_LifeTime, BP_MatPar, BP_TrCoord };

#define dble double
#define Buff_Size 65535

#ifndef min
#define min(a,b) (a>b?b:a)
#endif

#ifndef max
#define max(a,b) (a<b?b:a)
#endif

long line_length;	/* length of a line in bytes from header */
long *start_pos;	/* position of the first line in the stream */
long line_num;		/* number of lines */
long line_max;		/* maximal number of lines */
int  file_type;		/* type of datafile currently 4 or 5 */
int long_prec;
char *outfile;
FILE *out;
int  output_type;
int noeol;
int split;
/* 0-3: col, line, no col, no line */
#define CELLSIZE 100000	/* maximum number of ranges that can be given */
cellstr cells[4][CELLSIZE];
long cellp[4];		/* pointer to all four of the cellstr */
#define C_COL		0
#define C_LINE		1
#define C_NOCOL		2	/* disabled for type 4 */
#define C_NOLINE	3
#define C_PARAM		4	/* just for ToNumber */

/* array for objects */
#define OBJMAX 100
int obj_typ[OBJMAX];	/* objects type */
int *obj_num[OBJMAX];	/* objects number of this type */
int *obj_pos[OBJMAX];	/* their position in the file (obj. head not inc.) */
int obj_size[OBJMAX];	/* size of a single object */
int obj_col[OBJMAX];	/* number of columns that an object has */
int *obj_colt[OBJMAX];	/* type of the columns to be allocated */
			/* BP info included, int:1, double:2 */
int objp;		/* number of object groups */
int obj_timeall;	/* size of num and pos if line_length varies */
int *col_num;		/* size of line  in columns */

#define D_B	0
#define D_I	1
#define D_D	2
int data_size[3];	/* size of thses data types=sizeof(...) */
int dd,ii;

/* print out error message and stop */
void perr(char *format, ...)
{
  va_list arglist;

  va_start(arglist,format);
  vfprintf(stderr,format,arglist);
  fprintf(stderr,"\n");
  va_end(arglist);
  exit(0);
}

/* translate type information */
void Det_Obj_Params(int typ,int *size,int *col,int **colt)
{
  int t,a;

  *col=1;	/* everything has at least one double size */
  for (a=0;a<BP_N;a++) {
    if (typ&par_typ[a]) *col+=3;
  }
  if (typ&BP_Coord) *col-=2;
  if (typ&BP_Handle) *col-=2;
  if (typ&BP_LifeTime) *col-=2;
  if (typ&BP_TrCoord) *col-=2;
  if (typ&BP_Orient) *col+=1;
  *size=(*col-(((typ&BP_Coord)>0)+((typ&BP_Handle)>0)))*dd+
    		ii*(((typ&BP_Coord)>0)+((typ&BP_Handle)>0));
  t=typ&BP_Mask;
  if ((t==BT_Wall) || (t==BT_Cylinder) || (t==BT_Arrow)) {
    *col+=1;*size+=dd;
  }
  if ((t==BT_Force) || (t==BT_DTEdge)) {
    *col+=3;*size+=ii*2+dd;
  }
  if (t==BT_BallCC) {
    *col+=1;*size+=ii;
  }
  if ((t==BT_Ellipsoid) || (t==BT_Cone) || (t==BT_Triangle) ||
      (t==BT_Triangle2)) {
    *col+=2;*size+=2*dd;
  }
  /* put information in colt */
  (*colt)=(int *)malloc(ii*(*col));
  /* size */
  a=0;
  (*colt)[a++]=D_D;
  if ((t==BT_Ellipsoid) || (t==BT_Cone) || (t==BT_Triangle) ||
      (t==BT_Triangle2)) {
    (*colt)[a++]=D_D;
    (*colt)[a++]=D_D;
  }
  if (t==BT_BallCC) (*colt)[a++]=D_I;
  if ((t==BT_Wall) || (t==BT_Cylinder) || (t==BT_Arrow)) {
    (*colt)[a++]=D_D;
  }
  if ((t==BT_Force) || (t==BT_DTEdge)) {
    (*colt)[a++]=D_D;
    (*colt)[a++]=D_I;
    (*colt)[a++]=D_I;
  }
  for (t=0;t<BP_N;t++) {
    if (typ&par_typ[t]) {
      if ((par_typ[t]!=BP_Coord) && (par_typ[t]!=BP_Handle)) {
	(*colt)[a++]=D_D+par_typ[t];
	if ((par_typ[t]!=BP_LifeTime) && (par_typ[t]!=BP_TrCoord)) {
	  (*colt)[a++]=D_D+par_typ[t];
	  (*colt)[a++]=D_D+par_typ[t];
	}
	if (par_typ[t]==BP_Orient) (*colt)[a++]=D_D+par_typ[t];
      } else (*colt)[a++]=D_I+par_typ[t];
    }
  }
  if (a!=*col) perr("Internal error in Det_Obj_Params");
}

/* check outfile is set */
void Outfile(char **para,int np)
{
  int a;

  outfile=NULL;
  for (a=1;a<np;a++) {
    if (strcmp(para[a],"-o")==0) {
      a++;
      outfile=strdup(para[a]);
    }
    if (strcmp(para[a],"-ro")==0) {
      output_type=1;
    }
  }
}

/* check if line_max is set */
void Line_Max(char **para,int np)
{
  int a;

  line_max=-1;
  for (a=1;a<np;a++) {
    if (strcmp(para[a],"-lm")==0) {
      a++;line_max=atoi(para[a]);
    }
  }
}

/* analyze line: determine positions and type of each column */
void Analyze_Line(FILE *in,int mode)
{
  int typ,num,t,op;

  if (line_length>0) {
    obj_timeall=1;
  } else {
    obj_timeall=10000;
  }
  col_num=(int *) malloc(ii*obj_timeall);
  start_pos=(long *) malloc(sizeof(long)*obj_timeall);
  col_num[0]=1;
  objp=0;
  /* skip time */
  start_pos[0]=ftell(in);
  fseek(in,start_pos[0]+dd,SEEK_SET);
  do {
    fread(&typ,ii,1,in);
    /* typ=0 at the end of the line */
    if ((typ>0) && (objp<OBJMAX)) {
      obj_num[objp]=(int *) malloc(ii*obj_timeall);
      obj_pos[objp]=(int *) malloc(ii*obj_timeall);
      /* skip texture */
      if (file_type==4) fgetc(in);
      else fread(&num,ii,1,in);
      fread(&num,ii,1,in);
      obj_typ[objp]=typ;
      obj_num[objp][0]=num;
      obj_pos[objp][0]=ftell(in);
      Det_Obj_Params(typ,&obj_size[objp],&obj_col[objp],&obj_colt[objp]);
      col_num[0]+=obj_col[objp]*num+3;
      fseek(in,num*obj_size[objp],SEEK_CUR);
      objp++;
    } else if (objp>=OBJMAX) perr("too many objects (>%d)",OBJMAX);
  } while (typ>0);
  if (mode) return;
  data_size[0]=1;
  data_size[1]=ii;
  data_size[2]=dd;
  if (line_length<=0) {
    /* line_length is not fixed */
    t=1;	/* time */
    do {
      if ((line_max>0) && (t>line_max)) break;
      col_num[t]=1;
      start_pos[t]=ftell(in);
      fseek(in,dd,SEEK_CUR);
      if (feof(in)) break;
      op=0;
      do {
	fread(&typ,ii,1,in);
	if (typ==0) t++;
	if (feof(in)) break;
	/* typ=0 at the end of the line */
	if ((typ>0) && (op<objp)) {
	  /* skip texture */
	  if (file_type==4) fgetc(in);
	  else fread(&num,ii,1,in);
	  fread(&num,ii,1,in);
	  while ((op<objp) && (obj_typ[op]!=typ)) {
	    obj_num[op][t]=0;
	    op++;
	  }
	  if (op==objp) perr("Objects must keep order");
	  obj_num[op][t]=num;
	  obj_pos[op][t]=ftell(in);
	  col_num[t]+=obj_col[op]*num+3;
	  fseek(in,num*obj_size[op],SEEK_CUR);
	  op++;
	} else if (typ>0) perr("Cannot have new objects in time!");
      } while (typ>0);
      if (t>=obj_timeall) {
	obj_timeall*=2;
	realloc(obj_num[objp],ii*obj_timeall);
	realloc(obj_pos[objp],ii*obj_timeall);
	realloc(col_num,ii*obj_timeall);
	realloc(start_pos,sizeof(long)*obj_timeall);
      }
    } while (!feof(in));
    line_num=t-2;
  } else {
    fseek(in,0L,SEEK_END);
    line_num=(ftell(in)-start_pos[0])/line_length-1;
  }
}

void Print_File_Info(void)
{
  const char *obj_name[]= { "EOL(0)", "Wall(1)", "Ball(2)", "Cylinder(3)",
    			    "Disk(4)", "BallCC(5)", "Force(6)",
			    "STBall(7)", "DTEdge(8)", "DGBall(9)",
  			    "Ellipsoid(10)", "", "Arrow(12)", "Cone(13)",
  			    "Triangle(14)", "Triangle2(15)" };
  const char *par_name[]= { "Pos(x,y,z)", "Orient(rx,ry,rz,phi)",
    			    "Coord(Z)", "Velo(vx,vy,vz)",
			    "Force(fx,fy,fz)", "AngVelo(omx,omy,omz)",
			    "Torque(tx,ty,tz)", "TrVelo(tvx,tvy,tvz)",
			    "TrForce(tfx,tfy,tfz)",
			    "TrAngVelo(tox,toy,toz)",
  			    "Handle(hand)", "Lifetime(life)",
  			    "MatPar(mu,im,iI)", "TrCoord(tZ)" };
  int a,b;

  for (a=0;a<objp;a++) {
    printf("%d %s %d\n",a+1,
	obj_name[obj_typ[a]&BP_Mask],obj_num[a][0]);
    for (b=0;b<BP_N;b++) {
      if (obj_typ[a]&par_typ[b]) {
	printf("\t%s\n",par_name[b-1]);
      }
    }
  }
  exit(0);
}

/* convert a string to number +sign in front of number means from the
 * end */
long ToNumber(char *s,int c)
{
  int a,b,number,end;
  static char params[BP_N][4][5] = { { "dx", "dy", "dz", "i2" },
  				     { "x", "y", "z", "" },
    				     { "rx", "ry", "rz", "phi" },
				     { "Z", "", "", "" },
				     { "vx", "vy", "vz", "" },
				     { "fx", "fy", "fz", "" },
				     { "omx", "omy", "omz", "" },
				     { "tx", "ty", "tz", "" },
				     { "tvx", "tvy", "tvz", "" },
				     { "tfx", "tfy", "tfz", "" },
				     { "tox", "toy", "toz", "" },
  				     { "hand", "", "", "" },
  				     { "life", "", "", "" },
  				     { "mu", "im", "iI", "" },
  				     { "tZ", "", "", "" }};

  end=1;
  if (*s=='+') {
    end= -1;s++;
  }
  if ((c!=C_PARAM) || (isdigit(s[0]))) {
    if ((c==C_LINE) || (c==C_NOLINE)) {
      if (end==1) number=atoi(s);
      else number=line_num-atoi(s);
    } else {
      number=(atoi(s)+(end==-1))*end;
    }
  } else { /* C_PARAM and string */
    number=0;
    if (end<0) return(par_typ[BP_N-1]*2);
    for (a=0;a<BP_N;a++) {
      for (b=0;b<4;b++) {
	if (strlen(params[a][b])==0) break;
	if (strcmp(params[a][b],s)==0) number=b+par_typ[a];
      }
      if ((a==0) && (strcmp(s,"i1")==0)) number=2;
    }
  }
  return(number);
}

/* determine reange from a `from-to.inc' structure except from `from'
 * other fields may be missing */
void Range(char *s,int c,long *from,long *to,long *inc)
{
  int end;
  char *sp,*spp;

  *inc=1;
  if (strlen(s)==0) {
    *from=0;
    *to=ToNumber("+",c);
    return;
  }
  sp=s;
  end=0;
  spp=strchr(sp,'-');
  if (spp==NULL) {
    spp=sp+strlen(sp);
    end=1;
  }
  *spp='\0'; *from=ToNumber(sp,c);
  if (!end) {
    /* we have `*to' field */
    *spp='-';
    sp=spp+1;
    spp=strchr(sp,'.');
    if (spp==NULL) {
      spp=sp+strlen(sp);
      end=1;
    }
    *spp='\0';
    if (*sp=='\0') *to=ToNumber("+",c);
    else *to=ToNumber(sp,c);
    if (!end) {
      /* we have `*inc' field */
      *spp='.';
      spp++;
      *inc=ToNumber(spp,c);
    }
  } else *to=*from;
}

/* understand cmd line options */
void Translate_para(char **para,int np)
{
  int a,c,type,b;
  int parp;
  long from,to,inc;
  char *sp,*spnum,*sppar,ures[10];
  fields *f;

  c=0;
  for (a=2;a<np;a++) {
    if (strcmp(para[a],"-o")==0) {
      a++;
    } else if (strcmp(para[a],"-ro")==0) {
    } else if (strcmp(para[a],"-noeol")==0) {
      /* undocumented option to remove eol */
      noeol=1;
    } else if (strcmp(para[a],"-s")==0) {
      split=1;
    } else if (strcmp(para[a],"-p")==0) {
      long_prec=atoi(para[++a]);
    } else if (strcmp(para[a],"-c")==0) {
      c=0;
    } else if (strcmp(para[a],"-l")==0) {
      c=1;
    } else if (strcmp(para[a],"-nl")==0) {
      c=3;
    } else if (strcmp(para[a],"-lm")==0) {
      a++;
    } else {
      if (cellp[c]>CELLSIZE) {
	perr("Insufficient CELLSIZE (%d)\n",CELLSIZE);
	return;
      }
      if (para[a][0]=='%') {
	/* object */
	c=0;
	parp=0;
	ures[0]='\0';
	spnum=sppar=ures;
	sp=&para[a][1];
	/* object determination %a%b%c 
	 * `a': object type (number)
	 * `b': object num (range)
	 * `c': object param (range) */
	if ((*sp)=='g') {
	  sp++;
	  type=-atoi(sp);
	} else {
	  type=atoi(sp);
	}
	if (*sp!='%') sp=strchr(sp,'%');
	if (sp!=NULL) {
	  sp++;spnum=sp;
	  if (*sp!='%') sp=strchr(sp,'%');
	  if (sp!=NULL) {
	    *sp='\0';
	    sp++;sppar=sp;
	  }
	}
	/* count the number of fields */
	b=1; sp=spnum;
	while (*sp) {
	  if (*sp==',') b++;
	  sp++;
	}
	/* allocate memory for the field struct */
	f=(fields *)malloc(sizeof(fields));
	f->from=(int *)malloc(b*ii);
	f->to=(int *)malloc(b*ii);
	f->inc=(int *)malloc(b*ii);
	f->num=0;
	/* fill the fields */
	sppar=strtok(sppar,",");
	if (sppar==NULL) sppar=ures;
	while (sppar) {
	  Range(sppar,C_PARAM,&from,&to,&inc);
	  f->from[f->num]=from;
	  f->to[f->num]=to;
	  f->inc[f->num]=inc;
	  f->num++;
	  sppar=strtok(NULL,",");
	}
	/* get object number */
	spnum=strtok(spnum,",");
	if (spnum==NULL) spnum=ures;
	while (spnum) {
	  Range(spnum,C_COL,&from,&to,&inc);
	  cells[c][cellp[c]].from=from;
	  cells[c][cellp[c]].to=to;
	  cells[c][cellp[c]].inc=inc;
	  cells[c][cellp[c]].f=f;
	  cells[c][cellp[c]].type=type;
	  cellp[c]++;
	  spnum=strtok(NULL,",");
	}
      } else {
	/* not an object */
	sp=para[a];
	sp=strtok(sp,",");
	while (sp) {
	  Range(sp,c,&from,&to,&inc);
	  cells[c][cellp[c]].type=BP_Mask;
	  cells[c][cellp[c]].from=from;
	  cells[c][cellp[c]].to=to;
	  cells[c][cellp[c]].inc=inc;
	  cellp[c]++;
	  sp=strtok(NULL,",");
	}
      }
    }
  }
}

/* and line ranges are checked with a for cycle nl arguments checked
 * for every available field, not too slow, but any better idea?
 * XXX */
int Not_In(long l,int c)
{
  int a,rv;

  rv=1;
  for (a=0;a<cellp[c];a++) {
    if ((l>=cells[c][a].from) && (l<=cells[c][a].to) &&
	((l-cells[c][a].from)%cells[c][a].inc==0)) {
      return(0);
    }
  }
  return(1);
}

/* find file position of column c at time t */
long Find_Col(int t,int c,int *dt)
{
  int pos,a,b,p;

  pos=start_pos[t];
  /* well it's time, easy job */
  *dt=D_D;
  if (c==0) return(pos);
  /* Skip time from loop */
  c--;pos+=dd;
  *dt=D_I;
  p=0;			/* local object pointer */
  while ((c>0) && (p<objp)) {
    c--;pos+=ii;	/* type (int) */
    *dt=D_B;
    if (c==0) break;
    c--;pos++;		/* texture (byte) */
    *dt=D_I;
    if (c==0) break;
    c--;pos+=ii;	/* number of objects (int) */
    if (c<obj_num[p][t]*obj_col[p]) {
      /* c is whithin this object */
      a=c/obj_col[p];	/* a holds the index of the object c is in */
      c-=a*obj_col[p];
      pos+=a*obj_size[p];
      b=0;
      *dt=obj_colt[p][b]&BP_Mask;
      while (c>0) {
	c--;
	b++;
	pos+=data_size[*dt];
	*dt=obj_colt[p][b]&BP_Mask;
      }
      return(pos);
    }
    c-=obj_num[p][t]*obj_col[p];
    pos+=obj_num[p][t]*obj_size[p];
    p++;
  }
  if (p<objp) return(pos);
  else return(-1);
}

int Get_Pos(int col,int op,int ft)
{
  int a,c,t;

  t=obj_typ[op]&BP_Mask;
  if (col<BP_Mask) {
    /* size parameter */
    /* these bellow have only one parameter */
/*    if (((t==BT_Wall) || (t==BT_Disk) || (t==BT_STBall)) && (col>0)) {
      return(-1);
    }*/
    /* the rest has two except for BT_Force and BT_DTEdge */
/*    if ((col>1) && (t!=BT_Force) && (t!=BT_DTEdge)) return(-1);*/
    /* BT_Force has four parameters */
/*    if (col>3) return(-1);*/
    return(col);
  } else {
    /* optional parameter */
    c=col&BP_UpMask;
    if (ft) {
      for (a=0;a<obj_col[op];a++) {
	if (c<=(obj_colt[op][a]&BP_UpMask)) break;
      }
      if ((a==obj_col[op]) || (c<(obj_colt[op][a]&BP_UpMask))) a--;
    } else {
      for (a=obj_col[op]-1;a>=0;a--) {
	if (c>=(obj_colt[op][a]&BP_UpMask)) {
	  do {
	    a--;
	  } while ((a>=0) && (c==(obj_colt[op][a]&BP_UpMask)));
	  a++;
	  break;
	}
      }
      if ((a==-1) || (c>(obj_colt[op][a]&BP_UpMask))) a++;
    }
    /* parameter groups contain more columns */
    if (a<obj_col[op]-1) a+=(col&BP_Mask);
    if (a>=obj_col[op]) return(-1);	/* not that many column */
    return(a);
  }
}

void Print_ptr(void *ptr,int dt,FILE *out)
{
  if (output_type) {
    switch (dt) {
      case D_D:
	fwrite(ptr,dd,1,out);
	break;
      case D_I:
	fwrite(ptr,ii,1,out);
	break;
      case D_B:
	fwrite(ptr,sizeof(char),1,out);
	break;
    }
  } else {
    switch (dt) {
      case D_D:
	fprintf(out,"%.*le ",long_prec,*((double *)ptr));
	break;
      case D_I:
	fprintf(out,"%d ",*((int *)ptr));
	break;
      case D_B:
	fprintf(out,"%d ",*((char *)ptr));
	break;
    }
  }
}

/* print data according to determined options */
void Print_Data(FILE *in)
{
  int a,b,l,c,t,dt,d,e,f,g,po;
  long from,to,inc,cto,mm;
  long lp,fp,ir;
  void *ptr;

  if (cellp[C_LINE]==0) {
    /* no column is specified, then print all */
    cells[C_LINE][0].from=0;
    cells[C_LINE][0].to=line_num;
    cells[C_LINE][0].inc=1;
    cellp[C_LINE]=1;
  }
  if (cellp[C_COL]==0) {
    /* no column is specified, then print only time */
    cells[C_COL][0].from=0;
    cells[C_COL][0].to=0;
    cells[C_COL][0].inc=1;
    cellp[C_COL]=1;
  }
  ptr=(void *)malloc(Buff_Size);	/* XXX carefull with this */
  for (a=0;a<cellp[C_LINE];a++) {
    /* line ranges */
    for (l=cells[C_LINE][a].from; l<=min(cells[C_LINE][a].to,line_num);
	 l+=cells[C_LINE][a].inc) {
      if (Not_In(l,C_NOLINE)) {
	/* go to start of the line */
	if (line_length<=0) {
	  lp=start_pos[l];t=l;
	} else {
	  lp=start_pos[0]+l*line_length;t=0;
	}
	if (output_type) {
	  if (line_length<=0) {
	    if (l+1>=line_num) {
	      fseek(in,0,SEEK_END);
	      fp=ftell(in);
	    } else {
	      fp=start_pos[l+1];
	    }
	  } else {
	    fp=start_pos[0]+(l+1)*line_length;
	  }
	  fseek(in,lp,SEEK_SET);
	  fp-=lp; /* length */
	  if (noeol) fp-=sizeof(int);
	  while (fp>0) {
	    ir=fread(ptr,1,min(Buff_Size,fp),in);
	    fwrite(ptr,1,ir,out);
	    fp-=ir;
	  }
	} else {
	  /* columns */
	  for (b=0;b<cellp[C_COL];b++) {
	    cto=cells[C_COL][b].to;
	    if (cto<0) {
	      switch (cells[C_COL][b].type) {
		case BP_Mask:	/* absolute column */
		  cto=col_num[t]+1-cto;
		  break;
		case 0:		/* all objects */
		  mm=0;
		  for (d=0;d<objp;d++) {
		    if (mm<obj_num[d][t]) mm=obj_num[d][t];
		  }
		  cto=mm+1-cto;
		  break;
		default:
		  for (d=0;d<objp;d++) {
		    if ((cells[C_COL][b].type==(obj_typ[d]&BP_Mask)) ||
			((cells[C_COL][b].type<0) &&
			 (-1-cells[C_COL][b].type==d))) {
		      cto=obj_num[d][t]+1-cto;
		    }
		  }
		break;
	      }
	    }
	    for (c=cells[C_COL][b].from; c<=cto;
		 c+=cells[C_COL][b].inc) {
	      /* go to pos */
	      if (cells[C_COL][b].type==BP_Mask) {
		/* absolute column */
		fp=Find_Col(t,c,&dt);
		if (fp>0) {
		  fseek(in,fp,SEEK_SET);
		  fread(ptr,data_size[dt],1,in);
		  Print_ptr(ptr,dt,out);
		}
	      } else {
		/* object */
		for (d=0;d<objp;d++) 
		if ((cells[C_COL][b].type==0) ||
		    (cells[C_COL][b].type==(obj_typ[d]&BP_Mask)) ||
		    ((cells[C_COL][b].type<0) &&
		     (-1-cells[C_COL][b].type==d))) {
		  po=0;
		  for (e=0;e<cells[C_COL][b].f->num;e++) {
		    from=Get_Pos(cells[C_COL][b].f->from[e],d,0);
		    to=Get_Pos(cells[C_COL][b].f->to[e],d,1);
		    inc=cells[C_COL][b].f->inc[e];
		    if ((from>=0) && (to>=from) && (c<obj_num[d][t])) {
		      fp=obj_pos[d][t]+c*obj_size[d];
		      for (f=0;f<from;f++) {
			fp+=data_size[obj_colt[d][f]&BP_Mask];
		      }
		      for (f=from;f<=to;f+=inc) {
			dt=data_size[obj_colt[d][f]&BP_Mask];
			fseek(in,fp,SEEK_SET);
			fread(ptr,dt,1,in);
			Print_ptr(ptr,obj_colt[d][f]&BP_Mask,out);
			po=1;
			for (g=0;g<inc;g++) {
			  fp+=data_size[obj_colt[d][f+g]&BP_Mask];
			}
		      }
		    }
		  }
		  if ((po) && (split)) fprintf(out,"\n");
		}
	      }
	    }
	  }
	  if (!split) fprintf(out,"\n");
	}
      }
    }
  }
  if (outfile) fclose(out);
}

int main(np,para)
char *para[];
{
  int a,list;
  FILE *in;
  char s[100];

  if (np<2) {
    fprintf(stderr,"usage: getcol4 input [-p prec] [-lm num] [-s] [-ro] ... ");
    fprintf(stderr,"[-c] ... [-l] ... [-nl] ...\n");
    fprintf(stderr,"\t-o outfile = print output to file\n");
    fprintf(stderr,"\t-lm num = max number of lines to be read. ");
    fprintf(stderr,"Good for broken files!\n");
    fprintf(stderr,"\t-ro = binary report output. Normal output is text.\n");
    fprintf(stderr,"\t... = number or %%type%%num%%par   type:\n");
    fprintf(stderr,"\t1 Wall  \t2 Ball  \t3 Cylinder\n");
    fprintf(stderr,"\t4 Disk  \t5 BallCC\t6 Force\n");
    fprintf(stderr,"\t7 STBall\t8 DTEdge\t9 DGBall\n");
    fprintf(stderr,"\t10 Ellipsoid\t12 Arrow\t13 Cone\n");
    fprintf(stderr,"\t14 Triangle\n");
    exit(0);
  }
  if (np==2) list=1;
  else list=0;
  /* Open file */
  in=fopen(para[1],"rb");
  if (in==NULL) perr("input file (%s) cannot be open",para[1]);
  /* Header */
  file_type=0;
  output_type=0;
  long_prec=10;
  noeol=0;
  split=0;
  Outfile(para,np);
  if (outfile) {
    if ((out=fopen(outfile,"r"))) {
      fclose(out);
      perr("output file exists");
    }
    out=fopen(outfile,"wb");
  } else out=stdout;
  do {
    fgets(s,7,in);
    if (output_type) fputs(s,out);
    if (feof(in)) perr("Premature end of file or wrong file type");
    if (strcmp(s,"#type ")==0) {
      /* type */
      fscanf(in,"%d %d %d %ld", &file_type,&ii,&dd,&line_length);
      if (output_type)
	fprintf(out,"%d %d %d %ld", file_type,ii,dd,line_length);
      if ((file_type!=4) && (file_type!=6))
		perr("This code is for type 4 or 6 datafiles");
      if (ii!=sizeof(int)) perr("Size of int does not match");
      if (dd!=sizeof(double)) perr("Size of double does not match");
    }
  } while ((s[0]!='#') || (s[1]!='\n'));
  Line_Max(para,np);
  Analyze_Line(in,list);
  if (list) Print_File_Info();
  for (a=0;a<4;a++) {
    cellp[a]=0;
  }
  Translate_para(para,np);
  Print_Data(in);
  fclose(in);
  return(0);
}
