/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2005/01/25 18:41:57 $ $Revision: 1.57 $ */
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkgl.h>
#include <stdlib.h>
#include <stdio.h>
#include "global.h"
#include "rpn.h"
#include "body.h"
#include "object.h"
#include "XPM/icon.xpm"

/* prototypes */
void FrameOff(GtkWidget *,gpointer);
int RPN_decode(char *,RPN **,char **,int,int *,int,char [][RPN_StrLen]);
void RPN_Free(RPN **);
int ilog(double);
void Replot_Objects(int);
void Update_Constraint(GtkWidget*,gpointer);
int DoText(char *,constrglobal *,char **);
char *my_basename(char *);

const char *obj_name[16] = { "EOL", "Wall", "Ball", "Cylinder", "Disk",
			    "BallCC", "Force", "StBall", "DTEdge",
			    "DGBall", "Ellipsoid", "", "Arrow", "Cone", 
			    "Triangle","Triangle2" };

const char cef_parmainstr[CEF_PARNUM][10]={ "Num", "Base", "Pos",
  		"Orient", "Color", "Coordnum", "Velocity", "Force",
		"AngVelo", "Troque", "TrVelo", "TrForce", "TrAngVelo",
		"LifeTime", "MatPar", "TrCoord" };

#define CNN 12
const char constr_name[CNN][13]={"Arrow", "Color", "Cut", "Dir", "New",
  				"Position", "Texture", "Show", "Size",
				"Var", "Transparency", "Print" };
const int constr_parn[CNN]={	5, 6, 4, 3, 6,
				3, 3, 2, 3,
				4, 3, 6 };
const char constr_global_var[8][5]={ "x", "y", "z", "phix", "phiy",
  				"phiz", "Pos", "Ori" };

#define NTM "Not that many parameters"
const char constr_parstr[CNN][6][35]= {
  { "Direction(v)", "Length(s)", "Width(s)", "Expr", "Type", NTM },
  { "Red(s)/RGB(v)", "Green(s)/NULL", "Blue(s)/NULL", "Alpha(s)",
    "Expr", "Type" },
  { "Distance from origin(s)", "Width(s)", "Expr", "Type", NTM, NTM },
  { "Direction(v)", "Expr", "Type", NTM, NTM, NTM },
  { "Type(s)", "Expr(if empty only one is created)",
    	"Dimensions(v)", "Position(v)", "Direction", "Color(v)" },
  { "Position(v)", "Expr", "Type", NTM, NTM, NTM },
  { "Texture(s)", "Expr", "Type", NTM, NTM, NTM },
  { "Expr", "Type", NTM, NTM, NTM, NTM },
  { "Dimensions(v)", "Expr", "Type", NTM, NTM, NTM },
  { "Variable number(s)", "Value(s/v)", "Expr", "Type", NTM, NTM },
  { "Transparency(s)", "Expr", "Type", NTM, NTM, NTM },
  { "Text in c format eg: \"X: %%g\",x", "Position x(s)[0-1]",
    "Position y(s)[0-1]/Pos(v)", "Color(v)", "Angle(s)", "Size(s/v)" } };

void SetValueChangeText(void)
{
  char *s;
  double v;

  v=og.numchange;
  s=(char *)calloc(ilog(v)+20,sizeof(char));
  sprintf(s,"%lg",v);
  gtk_entry_set_text(GTK_ENTRY(og.valuechangeval), s);
  free(s);
}

int Move_Cut(double dy)
{
  int a,b,p,e;
  GtkTextIter start,end,iter,iter2;
  gchar *t,*s,*new;
  double v;

  gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
  gtk_text_buffer_get_iter_at_mark(og.tbuff,&iter,
      				gtk_text_buffer_get_insert(og.tbuff));
  s=gtk_text_buffer_get_text(og.tbuff,&start,&iter,FALSE);
  t=gtk_text_buffer_get_text(og.tbuff,&iter,&end,FALSE);
  if ((s[0]=='\0') && (t[0]=='\0')) {
    gtk_widget_hide(og.valuechangeval);
    gtk_widget_hide(og.valuechangetext);
    return(0);
  }
  p=e=0;
  a=strlen(s)-1;
  while (a>=0) {
    if ((e+p) && ((s[a]=='+') || (s[a]=='-'))) {
      a--;
      break;
    }
    if ((s[a]=='+') || (s[a]=='-')) {
      a--;
      if (a<1) break;
      if (tolower(s[a])!='e') break;
      if ((s[a-1]!='.') && (!isdigit(s[a-1]))) break;
    }
    if ((p) && (s[a]=='.')) break;
    if (s[a]=='.') p++;
    if (((e+p)) && (tolower(s[a])=='e')) break;
    if (tolower(s[a])=='e') e++;
    if ((!isdigit(s[a])) && (s[a]!='.') && (s[a]!='+') && (s[a]!='-') &&
	(tolower(s[a])!='e')) break;
    a--;
  }
  a++;
  if ((a==(int)strlen(s)) && (strlen(t)==0)) {
    gtk_widget_hide(og.valuechangeval);
    gtk_widget_hide(og.valuechangetext);
    return(0);
  }
  b=0;
  if ((a!=(int)strlen(s)-1) && ((t[0]=='-') || (t[0]=='+'))) b++;
  while (b<(int)strlen(t)) {
    if ((t[b]=='+') || (t[b]=='-')) {
      b++;
      if (b<1) break;
      if (tolower(t[b])!='e') break;
      if ((t[b-1]!='.') && (!isdigit(t[b-1]))) break;
    }
    if ((p+e) && (t[b]=='.')) break;
    if (t[b]=='.') p++;
    if ((e) && (tolower(t[b])=='e')) break;
    if (tolower(t[b])=='e') {
      if (b==(int)strlen(t)-1) break;
      if ((t[b+1]=='+') || (t[b+1]=='-')) b++;
      e++;
    }
    if ((!isdigit(t[b])) && (t[b]!='.') && (t[b]!='+') && (t[b]!='-') &&
	(tolower(t[b])!='e')) break;
    b++;
  }
  t[b]='\0';
  if ((a<(int)strlen(s)) || (b>0)) {
    new=(char *)calloc(b+strlen(s)-a+100,sizeof(char));
    sprintf(new,"%s%s",&s[a],t);
    v=atof(new);
    if (og.setnumoriginal==1) {
      og.numoriginal=v;
      og.setnumoriginal=0;
    }
    v+=dy*og.numchange;
    if (og.setnumoriginal==2) {
      v=og.numoriginal;
      og.setnumoriginal=0;
    }
    iter2=iter;
    gtk_text_iter_backward_chars(&iter,strlen(s)-a);
    gtk_text_iter_forward_chars(&iter2,b);
    gtk_text_buffer_delete(og.tbuff, &iter,&iter2);
    sprintf(new,"%lg",v);
    gtk_text_buffer_get_iter_at_mark(og.tbuff,&iter,
      				gtk_text_buffer_get_insert(og.tbuff));
    gtk_text_buffer_insert(og.tbuff,&iter,new,strlen(new));
    SetValueChangeText();
    Update_Constraint(NULL,(gpointer)1);
    free(new);
    return(1);
  } else {
    gtk_widget_hide(og.valuechangeval);
    gtk_widget_hide(og.valuechangetext);
    return(0);
  }
}

void ShowValueChange(int a)
{
  if (a) {
    gtk_widget_show(og.valuechangeval);
    gtk_widget_show(og.valuechangetext);
    SetValueChangeText();
  } else{
    gtk_widget_hide(og.valuechangeval);
    gtk_widget_hide(og.valuechangetext);
  }
}

void SetNumChange(int a)
{
  double factor[4]={ 1.1, 1.3, 2.0, 10.0 };

  if (a<=1) og.numchange=1.0;
  else {
    if (a>5) og.numchange*=factor[a-6];
    else og.numchange/=factor[5-a];
  }
  SetValueChangeText();
}

void SetOriginalNum(int set)
{
  og.setnumoriginal=set;
  if (set==2) Move_Cut(0.0);
}

void Set_fpos(int i)
{
  int b;

  for (b=0;b<bg.obj_num;b++) {
    body[b].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+2]=(double)i;
  }
}

void RPN_Dialog(char *s)
{
  GtkWidget *dialog,*label,*ok;

  dialog=gtk_dialog_new();
  label=gtk_label_new(s);
  ok=gtk_button_new_with_label("Ok");
  gtk_signal_connect_object (GTK_OBJECT (ok), "clicked",
		    GTK_SIGNAL_FUNC(gtk_widget_destroy), dialog);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),ok);
  gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),label);
  gtk_widget_show_all(dialog);
}

/* show/do not show object */
void Show_Onoff(GtkWidget *widget,gpointer data)
{
  bg.param[(int)data].show^=1;
}

/* Change between rad/deg */
void DegRad(GtkWidget *widget, gpointer data)
{
  int a,b;
  GtkWidget *dialog,*label,*ok;
  const char *dr[4]= {	"Radian unit is selected",
    			"Degrees unit is selected",
  			"New objects are not saved",
  			"New objects are saved" };

  a=(int)data;
  switch (a) {
    case 0:
      b=RPN_deg^=1;
      break;
    case 1:
      b=bg.save_new^=1;
      break;
  }
  dialog=gtk_dialog_new();
  label=gtk_label_new(dr[b+a*2]);
  ok=gtk_button_new_with_label("Ok");
  gtk_signal_connect_object (GTK_OBJECT (ok), "clicked",
		    GTK_SIGNAL_FUNC(gtk_widget_destroy), dialog);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),ok);
  gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),label);
  gtk_widget_show_all(dialog);
}

/* constraint error */

void Constraint_Error(char *s)
{
  gtk_label_set_label(GTK_LABEL(og.status),s);
}

void Free_Constraint(void)
{
  int a,b,c;

  for (b=0;b<bg.cgnum;b++) {
    for (a=0;a<CG_NP;a++) {
      if (cg[b].p[a]!=NULL) free(cg[b].p[a]);
    }
    if (cg[b].text!=NULL) free(cg[b].text);
    for (c=0;c<cg[b].textpnum;c++) {
      if (cg[b].textparam[c]!=NULL) free(cg[b].textparam[c]);
    }
    if ((cg[b].textpnum) && (cg[b].textparam!=NULL)) free(cg[b].textparam);
  }
  if (bg.cgnum) free(cg);
  bg.cgnum=0;
}

/* Update constraint */
void Update_Constraint(GtkWidget *widget, gpointer data)
{
  int a,b,parameters,i,replot,st;
  gchar *s,*sp,*spp,*sp2,*spmin,*p[CG_NP],*t;
  GtkTextIter start,end;

  anim.objectread=1;
  if (anim.constrreread>0) anim.constrreread=2;
  replot=(int)data;
  gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
  s=gtk_text_buffer_get_text(og.tbuff,&start,&end,FALSE);
  /* store in history */
  a=(og.hend+1)%CH_Max;
  if (a==og.hstart) {
    /* full */
    free(og.history[og.hstart]);
    og.hstart=(og.hstart+1)%CH_Max;
  }
  og.history[og.hend]=strdup(s);
  og.hindex=og.hend;
  og.hend=a;
  /* free previous constraints */
  Free_Constraint();
  /* count the constraints */
  b=-1;
  sp=s;
  do {
    spmin=sp+strlen(sp);
    for (a=0;a<CNN;a++) {
      spp=strstr(sp,constr_name[a]);
      if ((spp!=NULL) && (spp<spmin)) {
	spmin=spp;
      }
    }
    spp=sp;
    sp=spmin+1;
    b++;
  } while (spmin<spp+strlen(spp));
  /* allocate constraints */
  cg=(constrglobal *)calloc(b,sizeof(constrglobal));

  RPN_arrayfunc=0;
  sp=s;
  i=0;
  do {
    spmin=sp+strlen(sp);
    b=-1;
    for (a=0;a<CNN;a++) {
      spp=strstr(sp,constr_name[a]);
      if ((spp!=NULL) && (spp<spmin)) {
	spmin=spp;b=a;
      }
    }
    if (b>=0) {
      /* Constraint */
      spmin+=strlen(constr_name[b]);
      while ((*spmin!='\0') && (*spmin!='[')) spmin++;
      if (*spmin=='[') {
	/* Starting paranthese */
	spmin++;
	parameters=0;
	for (a=0;a<constr_parn[b];a++) {
	  spp=strchr(spmin,';');
	  sp2=strchr(spmin,']');
	  if (sp2==NULL) sp2=sp+strlen(sp);
	  if ((spp==NULL) || (sp2<spp)) {
	    /* we do not yet worry about lack of parameters */
	    spp=sp2;
	    if (*(spp+1)=='*') {
	      /* switch off constraint */
	      spmin=spp+2;
	      parameters=-1;
	      break;
	    }
	  }
	  if (spp!=NULL) {
	    *spp='\0';
	    p[a]=spmin;
	    spmin=spp+1;
	    parameters++;
	    if (spp==sp2) break;
	  } else p[a]=NULL;
	}
	if (parameters>0) {
	  /* analysis */
	  cg[i].type=b;
	  cg[i].text=NULL;
	  cg[i].textpnum=0;
	  for (a=0;a<CG_NP;a++) {
	    cg[i].p[a]=NULL;
	  }
	  st=0;
	  if (cg[i].type==9) {
	    if ((strlen(p[0])>3) && (strncmp(p[0],"gl.",3)==0)) {
	      for (a=0;a<8;a++) {
		if (strcmp(&p[0][3],constr_global_var[a])==0) break;
	      }
	      if (a==6) {
		spp=(char *)calloc(30+strlen(&p[0][3]),sizeof(char));
		t=strdup(spp);
		for (a=0;a<(int)strlen(t);a++) t[a]='_';
		gtk_label_set_label(GTK_LABEL(og.status),spp);
		gtk_label_set_pattern(GTK_LABEL(og.status),t);
		free(t);
		free(spp);
		Free_Constraint();
		return;
	      }
	      cg[i].p[0]=(RPN *)malloc(sizeof(RPN));
	      cg[i].p[0]->rpn=0;
	      cg[i].p[0]->type=RpnT_GlobalVar;
	      cg[i].p[0]->ival=a;
	      cg[i].p[0]->next=NULL;
	      st=1;
	    }
	  }
	  if (cg[i].type==11) {
	    cg[i].p[0]=(RPN *)malloc(sizeof(RPN));
	    cg[i].p[0]->rpn=0;
	    cg[i].p[0]->type=RpnT_Text;
	    cg[i].p[0]->next=NULL;
	    if (!DoText(p[0],&cg[i],&spp)) {
	      t=strdup(spp);
	      for (a=0;a<(int)strlen(t);a++) t[a]='_';
	      gtk_label_set_label(GTK_LABEL(og.status),spp);
	      gtk_label_set_pattern(GTK_LABEL(og.status),t);
	      free(t);
	      free(spp);
	      Free_Constraint();
	      return;
	    }
	    st=1;
	  }
	  for (a=st;a<parameters;a++) {
	    cg[i].arrayvar[a]=0;
	    /* check for global variables */
	    if (!RPN_decode(p[a],&cg[i].p[a],&spp,0,&cg[i].arrayvar[a],
			BG_RPN_globalvarnum,bg.rpnglobstr)) {
	      t=strdup(spp);
	      for (a=0;a<(int)strlen(t);a++) t[a]='_';
	      gtk_label_set_label(GTK_LABEL(og.status),spp);
	      gtk_label_set_pattern(GTK_LABEL(og.status),t);
	      free(t);
	      free(spp);
	      Free_Constraint();
	      return;
	    }
	  }
	  i++;
	  bg.cgnum++;
	}
      }
    }
    sp=spmin;
  } while (b>=0);
  if (replot) Replot_Objects(1);
}

/* cursor is moved */
void Constraint_Cursor(GtkWidget *widget, gpointer data)
{
  gchar *s,*sp,**kp,*t;
  GtkTextIter start,end,iter;
  int a,c,kpp,*kpt,d,cbc,cpbc,nfsc,doit;

  gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
  s=gtk_text_buffer_get_text(og.tbuff,&start,&end,FALSE);
  gtk_text_buffer_get_iter_at_mark(og.tbuff,&iter,
      				gtk_text_buffer_get_insert(og.tbuff));
  t=gtk_text_buffer_get_text(og.tbuff,&start,&iter,FALSE);
  cbc=strlen(t);
  d=strlen(s)-strlen(t);
  kp=(char **)calloc(strlen(s)/3,sizeof(char*));
  kpt=(int *)calloc(strlen(s)/3,sizeof(int));

  gtk_label_set_label(GTK_LABEL(og.status),"");
  kpp=0;
  for (a=0;a<CNN;a++) {
    sp=s;
    while ((sp=strstr(sp,constr_name[a]))) {
      kpt[kpp]=a;
      kp[kpp++]=sp;
      sp+=strlen(constr_name[a]);
    }
  }
  c=-1;
  for (a=0;a<kpp;a++) {
    /* previous */
    if ((kp[a]-s<cbc) && (kp[a]>kp[c])) {
      c=a;
    }
  }
  if (c>=0) {
    cpbc=kp[c]-s;
    nfsc=0;doit=0;
    for (a=cpbc;a<cbc;a++) {
      if (s[a]==';') nfsc++;
      if (s[a]=='[') doit=1;
      if (s[a]==']') doit=0;
    }
    if (doit) {
      if (nfsc>5) {
	/* Error */
	cbc=0;nfsc=5;
	kpt[c]=0;
      }
      gtk_label_set_label(GTK_LABEL(og.status),constr_parstr[kpt[c]][nfsc]);
    }
  }
  free(kp);
  free(kpt);
}

int Insert_String(char **s,int sp,int l,const char *t,int lt,int si)
{
  char *tmp;

  tmp=*s;
  /* overestimated size! */
  *s=(char *)calloc(lt+strlen(tmp),sizeof(char));
  strncpy(*s,tmp,sp);
  strncpy((*s)+sp,t,lt);
  strcpy((*s)+sp+lt,tmp+sp+l+si+1);
  return(1);
}

/* Character is inserted */
void Constraint_Insert(GtkWidget *widget, gpointer data)
{
  gchar *s,*sp,**kp,*psp,*t;
  const char *sc,*se;
  GtkTextIter start,end,iter;
  int a,b,c,kpp,*kpt,kppa,d,cfb,cbc,cpbc,l,frees,e;

  g_signal_handler_block(G_OBJECT (og.tbuff),og.constr_signalh);
  frees=0;
  gtk_label_set_label(GTK_LABEL(og.status),"");
  gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
  s=gtk_text_buffer_get_text(og.tbuff,&start,&end,FALSE);
  gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
  gtk_text_buffer_get_iter_at_mark(og.tbuff,&iter,
      				gtk_text_buffer_get_insert(og.tbuff));
  t=gtk_text_buffer_get_text(og.tbuff,&start,&iter,FALSE);
  cfb=strlen(t);
  d=strlen(s)-strlen(t);
  if (s[strlen(t)-1]=='\t') {
    if (og.istab) {
      sp=&s[strlen(t)-2-og.tabgoover];l=0;
    } else {
      sp=&s[strlen(t)-2];l=0;
    }
    while ((sp>s) && (isalnum(*sp))) {
      sp--;l++;
    }
    if ((sp==s) && (isalnum(*sp))) l++;
    else sp++;
    b=c=0;
    for (a=0;a<sp-s;a++) {
      b+=(s[a]=='[');
      c+=(s[a]==']');
    }
    if (b>c) {
      /* Inside constraint */
      b=0;se=sc=NULL;
      /* Unary functions */
      for (a=0;a<RPN_UONum;a++) {
	if (strncmp(sp,RPN_unaryoperator[a],l)==0) {
	  if (b==og.istab) sc=RPN_unaryoperator[a];
	  if (b==og.istab-1) se=RPN_unaryoperator[a];
	  b++;
	}
      }
      /* Vector functions */
      for (a=0;a<RPN_FNum;a++) {
	if (strncmp(sp,RPN_function[a],l)==0) {
	  if (b==og.istab) sc=RPN_function[a];
	  if (b==og.istab-1) se=RPN_function[a];
	  b++;
	}
      }
      /* Array functions */
      for (a=0;a<RPN_ArrFNum;a++) {
	if (strncmp(sp,RPN_arrayfunction[a],l)==0) {
	  if (b==og.istab) sc=RPN_arrayfunction[a];
	  if (b==og.istab-1) se=RPN_arrayfunction[a];
	  b++;
	}
      }
      /* Vectors */
      for (a=0;a<CEF_PARNUM;a++) {
	if ((strncmp(sp,RPN_parvectstr[a],l)==0) &&
	    (strlen(RPN_parvectstr[a])>0)) {
	  if (b==og.istab) sc=RPN_parvectstr[a];
	  if (b==og.istab-1) se=RPN_parvectstr[a];
	  b++;
	}
      }
      /* Scalar constants */
      for (a=0;a<RPN_ConstNum;a++) {
	if (strncmp(sp,RPN_const[a],l)==0) {
	  if (b==og.istab) sc=RPN_const[a];
	  if (b==og.istab-1) se=RPN_const[a];
	  b++;
	}
      }
      /* Vector constants */
      for (a=0;a<RPN_VConstNum;a++) {
	if (strncmp(sp,RPN_vconst[a],l)==0) {
	  if (b==og.istab) sc=RPN_vconst[a];
	  if (b==og.istab-1) se=RPN_vconst[a];
	  b++;
	}
      }

      if (se!=NULL) e=strlen(se)-l;
      else e=0;
      d-=e;
      if (sc!=0) {
	og.istab++;
	frees=Insert_String(&s,sp-s,l,sc,strlen(sc),e);
	og.tabgoover=strlen(sc)-l;
      } else {
	og.istab=0;
	frees=Insert_String(&s,sp-s,l,sp,l,e);
      }
    } else {
      /* Constraint string like Show */
      b=0;e=c=-1;
      for (a=0;a<CNN;a++) {
	if (strncmp(sp,constr_name[a],l)==0) {
	  if (b==og.istab) c=a;
	  if (b==og.istab-1) e=a;
	  b++;
	}
      }
      if (e>=0) e=strlen(constr_name[e])-l;
      else e=0;
      d-=e;
      if (c>=0) {
	og.istab++;
	frees=Insert_String(&s,sp-s,l,constr_name[c],
	    				strlen(constr_name[c]),e);
	og.tabgoover=strlen(constr_name[c])-l;
      } else {
	og.istab=0;
	frees=Insert_String(&s,sp-s,l,sp,l,e);
      }
    }
  } else og.istab=0;
  kp=(char **)calloc(strlen(s)/3,sizeof(char*));
  kpt=(int *)calloc(strlen(s)/3,sizeof(int));

  cbc=-1;
  kpp=0;
  for (a=0;a<CNN;a++) {
    sp=s;
    while ((sp=strstr(sp,constr_name[a]))) {
      kpt[kpp]=a;
      kp[kpp++]=sp;
      sp+=strlen(constr_name[a]);
    }
  }
  psp=s;
  gtk_text_buffer_delete(og.tbuff, &start,&end);
  for (a=0;a<kpp;a++) {
    sp=s+strlen(s);
    for (b=0;b<kpp-a;b++) {
      /* we do not need more sophisticated minimal search */
      if (kp[b]<sp) {
	c=b;sp=kp[b];
      }
    }
    kppa=kpp-a-1;
    sp=kp[c];kp[c]=kp[kppa];kp[kppa]=sp;
    b=kpt[c];kpt[c]=kpt[kppa];kpt[kppa]=b;
    if (sp-s<cfb) {
      cbc=kpt[c];
      cpbc=kp[c]-s;
    }
    if (sp-psp>0) {
      gtk_text_buffer_get_end_iter(og.tbuff, &end);
      gtk_text_buffer_insert(og.tbuff,&end,psp,sp-psp);
    } else {
      gtk_text_buffer_get_end_iter(og.tbuff, &end);
      gtk_text_buffer_insert(og.tbuff,&end," ",1);
    }
    gtk_text_buffer_get_end_iter(og.tbuff, &end);
    gtk_text_buffer_insert_with_tags_by_name(og.tbuff,&end,
	constr_name[b],-1,"bold",NULL);
    psp=sp+strlen(constr_name[b]);
  }
  gtk_text_buffer_get_end_iter(og.tbuff, &end);
  gtk_text_buffer_insert(og.tbuff,&end,psp,strlen(s)+s-psp);
  gtk_text_buffer_get_end_iter(og.tbuff, &end);
  gtk_text_iter_backward_chars(&end,d);
  gtk_text_buffer_place_cursor(og.tbuff,&end);
  free(kp);
  free(kpt);
  g_signal_handler_unblock(G_OBJECT (og.tbuff),og.constr_signalh);
  if (frees) free(s);
}

/* Constraint is inserted */
void Constraint_Insert_Text(GtkWidget *widget, gpointer data)
{
  GtkTextIter iter;
  char s[2]="[",*t;

  t=(char *)data;
  gtk_text_buffer_get_iter_at_mark(og.tbuff,&iter,
      				gtk_text_buffer_get_insert(og.tbuff));
  gtk_text_buffer_insert(og.tbuff,&iter,t, strlen(t));
  if ((t>=constr_name[0]) && (t<constr_name[0]+sizeof(constr_name)))  {
    gtk_text_buffer_get_iter_at_mark(og.tbuff,&iter,
      				gtk_text_buffer_get_insert(og.tbuff));
    gtk_text_buffer_insert(og.tbuff,&iter,s, 1);
  }
  gtk_widget_grab_focus(og.text);
}

void Constraint_Insert_Text2(GtkWidget *widget, GtkWidget *w,gpointer data)
{
  Constraint_Insert_Text(widget,data);
}

/* constraint undo/redo */
void Constraint_UnReDo(GtkWidget *widget, gpointer data)
{
  int a,b;
  GtkTextIter start,end;

  a=(int)data;
  switch (a) {
    case 0: /* undo */
      if ((og.hstart==og.hend) || (og.hindex==og.hstart)) break;
      og.hindex=(og.hindex+CH_Max-1)%CH_Max;
      gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
      gtk_text_buffer_delete(og.tbuff, &start,&end);
      gtk_text_buffer_get_end_iter(og.tbuff, &end);
      gtk_text_buffer_insert(og.tbuff,&end,og.history[og.hindex],
	  strlen(og.history[og.hindex]));
      break;
    case 1: /* redo */
      b=(og.hindex+1)%CH_Max;
      if ((og.hstart==og.hend) || (b==og.hend)) break;
      og.hindex=b;
      gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
      gtk_text_buffer_delete(og.tbuff, &start,&end);
      gtk_text_buffer_get_end_iter(og.tbuff, &end);
      gtk_text_buffer_insert(og.tbuff,&end,og.history[og.hindex],
	  strlen(og.history[og.hindex]));
      break;
  }
}

void Toggle_Widget(GtkWidget *widget,gpointer data)
{
  GtkWidget *w;

  w=(GtkWidget *)data;
  if (GTK_WIDGET_VISIBLE(w)) {
    gtk_widget_hide(w);
  } else {
    gtk_widget_show(w);
  }
}

void Create_Constant_Help(void)
{
  GtkStyle  *style;		/* style for pixmap background */
  GdkPixmap *pixmap;		/* pixmap for pixmap buttons */
  GdkBitmap *mask;		/* pixmap mask */
  GtkWidget *window;
  GtkWidget *table;
  GtkWidget *eventbox;
  GtkWidget *sw;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *label;
  GtkWidget *button;
  char s[256];
  int b,a;

  og.constanthelpisup=1;
  og.constanthelp=window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window),"Mozi constants");
  gtk_widget_realize(window);
  style= gtk_widget_get_style(window);
  pixmap = gdk_pixmap_create_from_xpm_d (window->window, &mask,
		       &style->bg[GTK_STATE_NORMAL],XPMicon);
  gdk_window_set_icon(window->window,NULL,pixmap,mask);
  gtk_container_set_border_width (GTK_CONTAINER (window), 0);
  g_signal_connect(window, "destroy",
		    G_CALLBACK (gtk_widget_hide), &window);
  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_set_size_request(sw,300,400);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
				       GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (window), sw);
  vbox=gtk_vbox_new(FALSE,0);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), vbox);
  gtk_widget_show(sw);
  table=gtk_table_new(2,RPN_ConstNum+RPN_VConstNum+3,FALSE);
  gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
  label=gtk_label_new("Name ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);
  gtk_widget_show(label);
  label=gtk_label_new("Value ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,0,1);
  gtk_widget_show(label);
  hbox=gtk_hbox_new(FALSE,0);
  gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,1,2);
  label=gtk_label_new_with_mnemonic("_S_c_a_l_a_r_:");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(hbox);
  gtk_widget_show(label);
  for (b=0;b<RPN_ConstNum;b++) {
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,b+2,b+3);
    eventbox=gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(hbox),eventbox,FALSE,FALSE,0);
    label=gtk_label_new(RPN_const[b]);
    gtk_container_add(GTK_CONTAINER(eventbox),label);
    gtk_widget_add_events(eventbox,GDK_BUTTON_PRESS_MASK);
    g_signal_connect(GTK_OBJECT (eventbox), "button_press_event",
		G_CALLBACK(Constraint_Insert_Text2),
		(gpointer)RPN_const[b]);
    gtk_widget_show(label);
    gtk_widget_show(eventbox);
    gtk_widget_show(hbox);
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,1,2,b+2,b+3);
    sprintf(s,"%lg",RPN_constval[b]);
    label=gtk_label_new(s);
    gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
  }
  hbox=gtk_hbox_new(FALSE,0);
  gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,b+2,b+3);
  label=gtk_label_new_with_mnemonic("_V_e_c_t_o_r_:");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(hbox);
  gtk_widget_show(label);
  a=RPN_ConstNum;
  for (b=0;b<RPN_VConstNum;b++) {
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,b+3+a,b+4+a);
    eventbox=gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(hbox),eventbox,FALSE,FALSE,0);
    label=gtk_label_new(RPN_vconst[b]);
    gtk_container_add(GTK_CONTAINER(eventbox),label);
    gtk_widget_add_events(eventbox,GDK_BUTTON_PRESS_MASK);
    g_signal_connect(GTK_OBJECT (eventbox), "button_press_event",
		G_CALLBACK(Constraint_Insert_Text2),
		(gpointer)RPN_vconst[b]);
    gtk_widget_show(label);
    gtk_widget_show(eventbox);
    gtk_widget_show(hbox);
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,1,2,b+3+a,b+4+a);
    sprintf(s,"(%lg,%lg,%lg)",RPN_vconstval[b][0],RPN_vconstval[b][1],
				RPN_vconstval[b][2]);
    label=gtk_label_new(s);
    gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
  }
  gtk_widget_show(table);
  button=gtk_button_new_with_label("Hide this window");
  gtk_box_pack_start(GTK_BOX(vbox),button,FALSE,FALSE,0);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
		    G_CALLBACK (gtk_widget_hide), (gpointer)window);
  gtk_widget_show(button);
  gtk_widget_show(vbox);
  gtk_widget_show(window);
}

void Create_Function_Help(void)
{
  GtkStyle  *style;		/* style for pixmap background */
  GdkPixmap *pixmap;		/* pixmap for pixmap buttons */
  GdkBitmap *mask;		/* pixmap mask */
  GtkWidget *window;
  GtkWidget *table;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *eventbox;
  GtkWidget *sw;
  GtkWidget *text;
  GtkTextBuffer *tbuff;
  int a,b;
  char s[8];
  const char variablehelptext[]="Mozi has scalar and vector variables. \
In the below tables the cases are seperated according to the nature of \
the argument or the result. The formula analyzer does not check for \
mistakes like taking the sin of a vector or a vector being in the \
denominator but in those cases simply takes the absolut value of the \
vector as a scalar and continues with that. On the other hand vector \
functions does not do anything with the scalars but return the first \
argument in a vector form of (s,0,0).";

  const char bindescr_text[RPN_BINNUM][50]={
    		"Power (s,*)/Vectorial product(v^v)",
                "Multiplication (s,*)/Scalar product(v*v)",
		"Division (s,s)",
		"Integer modulus (s,s)",
                "Addition v+s=v, s+v=s",
		"Substruction v-s=v,s-v=s",
		"=1 if greater or equal, =0 otherwise",
		"=1 if equal, =0 otherwise",
		"=1 if lesser or equal, =0 otherwise",
		"=1 if not equal, =0 otherwise",
		"=1 if greater, =0 otherwise",
		"=1 if lesser, =0 otherwise",
		"=1 if both are greater than 0, =0 otherwise",
		"=1 if any is greater than 0, =0 otherwise" };

  const char unidescr_text[RPN_UONum][50]={
	"sinus (s)", "cosinus (s)", "tangent (s)", "cotangent (s)",
	"arcussinus (s)", "arcuscosinus (s)", "arcustanget (s)",
	"arcuscotangent (s)", "square root of absolute value (s)",
	"square (s)", "absolut value (s)", "sign function (s)",
	"exponential (s)", "natural logarithm (s)", "integer part (s)",
	"sgn (s), normalization (v)", "value of variable arg: (s) rv: (s/v)",
  	"num of the first body of group (s)", "number of bodies in group (s)" };

  const char vectorparam[RPN_FNum][40]={ "r[v]=Rot(t[v],phi[s],x[v]) ",
	"r[v]=Mirror(t[v],x[v]) ", "r[v]=Lincomb(a[v],b[v]) ",
	"r[s/v]=If(e[s],t[s/v],f[s/v]) ", "r[v]=Vector(x[s],y[s],z[s]) ",
	"r[s]=Comp(x[v],c[s]) ", "r[v]=Projl(p[v],x[v]) ",
	"r[v]=Projp(p[v],x[v]) " };
  const char vecdescr_text[RPN_FNum][70]={
	"Rotates vector x around axis t with angle phi",
	"Mirrors vector x on the plane perpendicular to t",
	"Componentwise sum: r_x=a_x+b_x, r_y=a_y+b_y, r_z=a_z+b_z",
	"If e>0 then r=t otherwise r=f. t and f must be of the same kind",
	"Returns vector given by the three components",
	"Returns c component of vector. c must be 0, 1, 2",
	"Projects x on the vector p. The result is one dimensional",
	"Projects x on the plane perpendicular to p" 
  };
  const char arrdescr_text[RPN_ArrFNum][70]={
	"Minimum of arguments",
	"Maximum of arguments",
	"Average of arguments",
	"Sum of arguments",
	"num of minimum of arguments",
	"num of maximum of arguments",
	"returns num only if argument is true (>0)" };

  og.functionhelpisup=1;
  og.functionhelp=window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window),"Mozi functions");
  gtk_widget_realize(window);
  style= gtk_widget_get_style(window);
  pixmap = gdk_pixmap_create_from_xpm_d (window->window, &mask,
		       &style->bg[GTK_STATE_NORMAL],XPMicon);
  gdk_window_set_icon(window->window,NULL,pixmap,mask);
  gtk_container_set_border_width (GTK_CONTAINER (window), 0);
  g_signal_connect(window, "destroy",
		    G_CALLBACK (gtk_widget_hide), &window);
  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_set_size_request(sw,590,400);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
				       GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (window), sw);
  vbox=gtk_vbox_new(FALSE,0);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), vbox);
  gtk_widget_show(sw);
  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  label=gtk_label_new("Small help");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  button=gtk_check_button_new_with_label("Off");
  gtk_box_pack_end(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),TRUE);
  gtk_widget_show(label);
  gtk_widget_show(button);
  gtk_widget_show(hbox);
  tbuff=gtk_text_buffer_new(NULL);
  text=gtk_text_view_new_with_buffer(GTK_TEXT_BUFFER(tbuff));
  gtk_text_view_set_editable(GTK_TEXT_VIEW(text), FALSE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(text),GTK_WRAP_WORD);
  gtk_box_pack_start(GTK_BOX(vbox),text,FALSE,FALSE,0);
  gtk_text_buffer_set_text(GTK_TEXT_BUFFER(tbuff),
      		variablehelptext,strlen(variablehelptext));
  gtk_signal_connect(GTK_OBJECT(button),"toggled",
		GTK_SIGNAL_FUNC (Toggle_Widget), (gpointer)text);

  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  label=gtk_label_new_with_mnemonic("Basic _b_i_n_a_r_y functions");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  button=gtk_check_button_new_with_label("Off");
  gtk_box_pack_end(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),TRUE);
  gtk_widget_show(label);
  gtk_widget_show(button);
  gtk_widget_show(hbox);
  table=gtk_table_new(3,RPN_BINNUM+1,FALSE);
  gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
  gtk_signal_connect(GTK_OBJECT(button),"toggled",
		GTK_SIGNAL_FUNC (Toggle_Widget), (gpointer)table);
  label=gtk_label_new("Function ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);
  gtk_widget_show(label);
  label=gtk_label_new("Order ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,0,1);
  gtk_widget_show(label);
  label=gtk_label_new("Description");
  gtk_table_attach_defaults(GTK_TABLE(table),label,2,3,0,1);
  gtk_widget_show(label);
  for (a=1,b=0;b<RPN_BINNUM;a++,b++) {
    eventbox=gtk_event_box_new();
    gtk_table_attach_defaults(GTK_TABLE(table),eventbox,0,1,a,a+1);
    if (RPN_binaryoperator[b][0]==RPN_binaryoperator[b][1]) {
      sprintf(s,"%s, %s",RPN_binaryoperator[b],RPN_binaryoperator[b+1]);
      b++;
      label=gtk_label_new(s);
    } else {
      label=gtk_label_new(RPN_binaryoperator[b]);
    }
    gtk_container_add(GTK_CONTAINER(eventbox),label);
    gtk_widget_add_events(eventbox,GDK_BUTTON_PRESS_MASK);
    g_signal_connect(GTK_OBJECT (eventbox), "button_press_event",
		G_CALLBACK(Constraint_Insert_Text2),
		(gpointer)RPN_binaryoperator[b]);
    gtk_widget_show(label);
    gtk_widget_show(eventbox);
    sprintf(s,"%d",RPN_binoporder[b]+1);
    label=gtk_label_new(s);
    gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,a,a+1);
    gtk_widget_show(label);
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,2,3,a,a+1);
    label=gtk_label_new(bindescr_text[a-1]);
    gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
  }

  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  label=gtk_label_new_with_mnemonic("_U_n_a_r_y functions");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  button=gtk_check_button_new_with_label("Off");
  gtk_box_pack_end(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),TRUE);
  gtk_widget_show(label);
  gtk_widget_show(button);
  gtk_widget_show(hbox);
  table=gtk_table_new(2,RPN_BINNUM+1,FALSE);
  gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
  gtk_signal_connect(GTK_OBJECT(button),"toggled",
		GTK_SIGNAL_FUNC (Toggle_Widget), (gpointer)table);
  label=gtk_label_new("Function ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);
  gtk_widget_show(label);
  label=gtk_label_new("Description");
  gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,0,1);
  gtk_widget_show(label);
  for (b=0;b<RPN_UONum;b++) {
    eventbox=gtk_event_box_new();
    gtk_table_attach_defaults(GTK_TABLE(table),eventbox,0,1,b+1,b+2);
    label=gtk_label_new(RPN_unaryoperator[b]);
    gtk_container_add(GTK_CONTAINER(eventbox),label);
    gtk_widget_add_events(eventbox,GDK_BUTTON_PRESS_MASK);
    g_signal_connect(GTK_OBJECT (eventbox), "button_press_event",
		G_CALLBACK(Constraint_Insert_Text2),
		(gpointer)RPN_unaryoperator[b]);
    gtk_widget_show(label);
    gtk_widget_show(eventbox);
    gtk_widget_show(label);
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,1,2,b+1,b+2);
    label=gtk_label_new(unidescr_text[b]);
    gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
  }

  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  label=gtk_label_new_with_mnemonic("_V_e_c_t_o_r functions");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  button=gtk_check_button_new_with_label("Off");
  gtk_box_pack_end(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),TRUE);
  gtk_widget_show(label);
  gtk_widget_show(button);
  gtk_widget_show(hbox);
  table=gtk_table_new(3,RPN_BINNUM+1,FALSE);
  gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
  gtk_signal_connect(GTK_OBJECT(button),"toggled",
		GTK_SIGNAL_FUNC (Toggle_Widget), (gpointer)table);
  label=gtk_label_new("Function ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);
  gtk_widget_show(label);
  label=gtk_label_new("Parameter ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,0,1);
  gtk_widget_show(label);
  label=gtk_label_new("Description");
  gtk_table_attach_defaults(GTK_TABLE(table),label,2,3,0,1);
  gtk_widget_show(label);
  for (b=0;b<RPN_FNum;b++) {
    eventbox=gtk_event_box_new();
    gtk_table_attach_defaults(GTK_TABLE(table),eventbox,0,1,b+1,b+2);
    label=gtk_label_new(RPN_function[b]);
    gtk_container_add(GTK_CONTAINER(eventbox),label);
    gtk_widget_add_events(eventbox,GDK_BUTTON_PRESS_MASK);
    g_signal_connect(GTK_OBJECT (eventbox), "button_press_event",
		G_CALLBACK(Constraint_Insert_Text2),
		(gpointer)RPN_function[b]);
    gtk_widget_show(label);
    gtk_widget_show(eventbox);
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,1,2,b+1,b+2);
    label=gtk_label_new(vectorparam[b]);
    gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,2,3,b+1,b+2);
    label=gtk_label_new(vecdescr_text[b]);
    gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
  }

  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  label=gtk_label_new_with_mnemonic("_A_r_r_a_y functions");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  button=gtk_check_button_new_with_label("Off");
  gtk_box_pack_end(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),TRUE);
  gtk_widget_show(label);
  gtk_widget_show(button);
  gtk_widget_show(hbox);
  table=gtk_table_new(2,RPN_BINNUM+1,FALSE);
  gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
  gtk_signal_connect(GTK_OBJECT(button),"toggled",
		GTK_SIGNAL_FUNC (Toggle_Widget), (gpointer)table);
  label=gtk_label_new("See also help about variables!");
  gtk_table_attach_defaults(GTK_TABLE(table),label,0,2,0,1);
  gtk_widget_show(label);
  label=gtk_label_new("Function ");
  gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,1,2);
  gtk_widget_show(label);
  label=gtk_label_new("Description");
  gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,1,2);
  gtk_widget_show(label);
  for (b=0;b<RPN_ArrFNum;b++) {
    eventbox=gtk_event_box_new();
    gtk_table_attach_defaults(GTK_TABLE(table),eventbox,0,1,b+2,b+3);
    label=gtk_label_new(RPN_arrayfunction[b]);
    gtk_container_add(GTK_CONTAINER(eventbox),label);
    gtk_widget_add_events(eventbox,GDK_BUTTON_PRESS_MASK);
    g_signal_connect(GTK_OBJECT (eventbox), "button_press_event",
		G_CALLBACK(Constraint_Insert_Text2),
		(gpointer)RPN_arrayfunction[b]);
    gtk_widget_show(label);
    gtk_widget_show(eventbox);
    gtk_widget_show(label);
    hbox=gtk_hbox_new(FALSE,0);
    gtk_table_attach_defaults(GTK_TABLE(table),hbox,1,2,b+2,b+3);
    label=gtk_label_new(arrdescr_text[b]);
    gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
  }

  button=gtk_button_new_with_label("Hide this window");
  gtk_box_pack_start(GTK_BOX(vbox),button,FALSE,FALSE,0);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
		    G_CALLBACK (gtk_widget_hide), (gpointer)window);
  gtk_widget_show(button);
  gtk_widget_show(vbox);
  gtk_widget_show(window);
}

void Create_Variable_Help(int mode)
{
  GtkStyle  *style;		/* style for pixmap background */
  GdkPixmap *pixmap;		/* pixmap for pixmap buttons */
  GdkBitmap *mask;		/* pixmap mask */
  GtkWidget *window;
  GtkWidget *table;
  GtkWidget *hbox;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *sw;
  int a,b,c,d;
  char s[256];
  const char fieldname[CEF_PARNUM][30]={ "Number/type: ",
    "Size/force index: ", "Position: ", "Orientation: ", "Color: ",
    "Coordination number: ", "Velocity: ", "Force: ",
    "Angular velocity: ", "Torque: ", "Tr. Velocity: ", "Tr. Force: ",
    "Tr. Angular velocity: ", "LifeTime:", "Material Parameters: ",
    "Tr. Coordination number", "Time: " };
  const char fieldname2[CEF_PARNUM][30]={ "Position", "Orientation",
	"Coordination number", "Velocity", "Force",
	"Angular Velocity", "Torque", "Tr. Velocity", "Tr. Force",
	"Tr. Ang. Velocity", "Handle", "LifeTime", "Material Par.",
  	"Tr. Coord. num." };

  og.variablehelpisup|=mode;
  og.variablehelp=window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  if (mode==1) {
    gtk_window_set_title (GTK_WINDOW (window),"Mozi variables");
  } else {
    gtk_window_set_title (GTK_WINDOW (window),"Object parameters");
  }
  gtk_widget_realize(window);
  style= gtk_widget_get_style(window);
  pixmap = gdk_pixmap_create_from_xpm_d (window->window, &mask,
		       &style->bg[GTK_STATE_NORMAL],XPMicon);
  gdk_window_set_icon(window->window,NULL,pixmap,mask);
  gtk_container_set_border_width (GTK_CONTAINER (window), 0);
  g_signal_connect(window, "destroy",
		    G_CALLBACK (gtk_widget_hide), &window);
  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
				       GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (window), sw);
  if (mode==1) {
    gtk_widget_set_size_request(sw,400,440);
    table=gtk_table_new(4,CEF_PARNUM+1,FALSE);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), table);
    for (a=0;a<CEF_PARNUM;a++) {
      label=gtk_label_new(fieldname[a]);
      gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,a,a+1);
      hbox=gtk_hbox_new(FALSE,0);
      gtk_table_attach_defaults(GTK_TABLE(table),hbox,1,2,a,a+1);
      b=0;
      while ((strlen(RPN_parstr[a][b])) && (b<CEF_SUBPARNUM)) {
	button=gtk_button_new_with_label(RPN_parstr[a][b]);
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
	g_signal_connect(button, "clicked",
		      G_CALLBACK (Constraint_Insert_Text),
		      (gpointer)RPN_parstr[a][b]);
	b++;
      }
      if (strlen(RPN_parvectstr[a])) {
	label=gtk_label_new(" Vector: ");
	gtk_table_attach_defaults(GTK_TABLE(table),label,2,3,a,a+1);
	button=gtk_button_new_with_label(RPN_parvectstr[a]);
	gtk_table_attach_defaults(GTK_TABLE(table),button,3,4,a,a+1);
	g_signal_connect(button, "clicked",
		      G_CALLBACK (Constraint_Insert_Text),
		      (gpointer)RPN_parvectstr[a]);
      }
    }
  } else { /* mode==2 */
    gtk_widget_set_size_request(sw,200,100*max(1,min(bg.obj_num-1,5)));
    a=0;
    for (b=0;b<bg.obj_num;b++) {
      a++;
      for (c=BP_Pos;c<(BP_Pos<<(CEF_PARNUM-3));c<<=1) {
	if (bg.param[b].par&c) a++;
      }
    }
    table=gtk_table_new(2,a+3,FALSE);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), table);
    a=0;
    for (b=0;b<bg.obj_num;b++) {
      sprintf(s,"%s(%d)",obj_name[bg.param[b].type],b+1);
      label=gtk_label_new(s);
      gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,a,a+1);
      a++;d=0;
      for (c=BP_Pos;c<(BP_Pos<<(CEF_PARNUM-3));c<<=1,d++) {
	if (bg.param[b].par&c) {
	  label=gtk_label_new(fieldname2[d]);
	  gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,a,a+1);
	  a++;
	}
      }
    }
  }
  button=gtk_button_new_with_label("Hide this window");
  gtk_table_attach_defaults(GTK_TABLE(table),button,0,6-mode*2,a,a+1);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
		    G_CALLBACK (gtk_widget_hide), (gpointer)window);
  gtk_widget_show_all(window);
}

/* constraint frames show/off */
void Constraint_Part(GtkWidget *widget, gpointer data)
{
  int a;

  a=(int)data;
  switch (a) {
    case 0:
      if (GTK_WIDGET_VISIBLE(og.text)) {
	gtk_widget_hide(og.text);
	gtk_widget_hide(og.scroll);
	gtk_widget_hide(og.bhbox);
      } else {
	gtk_widget_show(og.text);
	gtk_widget_show(og.scroll);
	gtk_widget_show(og.bhbox);
      }
      break;
    case 1:
      if (GTK_WIDGET_VISIBLE(og.chbox[0])) {
	gtk_widget_hide(og.chbox[0]);
	gtk_widget_hide(og.chbox[1]);
      } else {
	gtk_widget_show(og.chbox[0]);
	gtk_widget_show(og.chbox[1]);
      }
      break;
    case 2:
      if (og.constanthelpisup==0) Create_Constant_Help();
      gtk_widget_show(og.constanthelp);
      break;
    case 3:
      if (og.functionhelpisup==0) Create_Function_Help();
      gtk_widget_show(og.functionhelp);
      break;
    case 4:
      if ((og.variablehelpisup&1)==0) Create_Variable_Help(1);
      gtk_widget_show(og.variablehelp);
      break;
    case 5:
      if ((og.variablehelpisup&2)==0) Create_Variable_Help(2);
      gtk_widget_show(og.variablehelp);
      break;
  }
}

void Load_Constraint_From_File(const char *filename,int tohist)
{
  GtkTextIter iter;
  FILE *cin;
  char s[256],*t;
  int a,b,doit;

  cin=fopen(filename,"r");
  if (cin==NULL) {
    fprintf(stderr,"Cannot open file %s for reading\n",filename);
  } else {
    if (tohist) {
      doit=1;
      for (a=og.fhstart;a!=og.fhend;a=(a+1)%CH_Max) {
	if (strcmp(og.fnamehist[a],filename)==0) {
	  doit=0;break;
	}
      }
      b=(og.fhend+CH_Max-1)%CH_Max;
      if (doit) {
	a=(og.fhend+1)%CH_Max;
	if (a==og.fhstart) {
	  /* full */
	  free(og.fnamehist[og.fhstart]);
	  og.fhstart=(og.fhstart+1)%CH_Max;
	}
	og.fnamehist[og.fhend]=strdup(filename);
	og.fhend=a;
      } else if (a!=b) {
	t=og.fnamehist[a];
	og.fnamehist[a]=og.fnamehist[b];
	og.fnamehist[b]=t;
      }
    }
    while (!feof(cin)) {
      fgets(s,256,cin);
      if (!feof(cin)) {
	gtk_text_buffer_get_iter_at_mark(og.tbuff,&iter,
      				gtk_text_buffer_get_insert(og.tbuff));
	gtk_text_buffer_insert(og.tbuff,&iter,s,strlen(s));
      }
    }
    fclose(cin);
  }
}

void Save_Constr_As(GtkWidget *widget, gpointer data)
{
  GtkWidget *filesel;
  GtkTextIter start,end;
  const char *s,*t;
  FILE *out;

  filesel=(GtkWidget*)data;
  s=gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
  t=gtk_text_buffer_get_text(og.tbuff,&start,&end,FALSE);
  out=fopen(s,"w");
  fputs(t,out);
  fclose(out);
}

void Insert_Constraint(GtkWidget *widget, gpointer data)
{
  GtkWidget *filesel;
  const char *s;

  filesel=(GtkWidget*)data;
  s=gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  Load_Constraint_From_File(s,1);
}

void Load_Constraint(GtkWidget *widget, gpointer data)
{
  GtkWidget *filesel;
  GtkTextIter start,end;
  const char *s;

  filesel=(GtkWidget*)data;
  s=gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
  gtk_text_buffer_delete(og.tbuff, &start,&end);
  Load_Constraint_From_File(s,1);
}

void Tree_Selection(GtkTreeSelection *sel,gpointer data)
{
  int a,type,l;
  char *s,*t;
  GtkTreeModel *model;
  GtkTextIter start,end,iter;
  GtkTreeIter titer;
  GtkTextBuffer *tb;
  FILE *scin;

  a=(int)data;
  switch (a/2) {
    case 0:
      tb=og.showtbuff;
      gtk_text_buffer_get_bounds(tb, &start,&end);
      gtk_text_buffer_delete(tb, &start,&end);
      break;
    case 1:
      tb=og.tbuff;
      gtk_text_buffer_get_bounds(tb, &start,&end);
      gtk_text_buffer_delete(tb, &start,&end);
      break;
    case 2:
      tb=og.tbuff;
      break;
  }
  if (gtk_tree_selection_get_selected (sel, &model, &titer)) {
    gtk_tree_model_get(model, &titer,0,&t,1,&s,2,&type,-1);
    free(t);
    gtk_text_buffer_get_iter_at_mark(tb,&iter,
				gtk_text_buffer_get_insert(tb));
    if (type) {
      scin=fopen(s,"r");
      if (scin==NULL) {
	t=calloc(strlen(s)+strlen("Cannot open file .  "),sizeof(char));
	sprintf(t,"Cannot open file %s.",s);
      } else {
	fseek(scin,0,SEEK_END);
	l=ftell(scin);
	rewind(scin);
	t=calloc(l+1,sizeof(char));
	fread(t,1,l,scin);
	fclose(scin);
      }
      gtk_text_buffer_insert(tb,&iter,t,strlen(t));
      free(t);
    } else {
      gtk_text_buffer_insert(tb,&iter,s,strlen(s));
    }
    free(s);
  }
}

void Remove_File_From_List(GtkWidget *widget, gpointer data)
{
  GtkTreeSelection *sel;
  GtkTreeModel *model;
  GtkTreeIter iter;
  char *t,*s;
  int type,a,b,doit;
  constrstr *p;

  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(og.treeview));
  if (gtk_tree_selection_get_selected (sel, &model, &iter)) {
    gtk_tree_model_get(model, &iter,0,&t,1,&s,2,&type,-1);
    gtk_list_store_remove(GTK_LIST_STORE (model), &iter);
    if (type) {
      /* file */
      doit=0;
      for (a=og.fhstart;a!=og.fhend;a=(a+1)%CH_Max) {
	if (strcmp(s,og.fnamehist[a])==0) {
	  free(og.fnamehist[a]);
	  doit=1;
	}
	if (doit) {
	  b=(a+1)%CH_Max;
	  og.fnamehist[a]=og.fnamehist[b];
	}
      }
      og.fhend=(og.fhend+CH_Max-1)%CH_Max;
      free(s);
    } else {
      /* named */
      free(s);
      named_constraintp=&named_constraint_start;
      p=NULL;
      while (named_constraintp!=NULL) {
	s=strdup((*named_constraintp)->s);
	s++;
	for (a=0;a<(int)strlen(s);a++) {
	  if (s[a]==127) {
	    s[a]='\0';
	    break;
	  }
	}
	if (strcmp(s,t)==0) {
	  if (p==NULL) {
	    p=named_constraint_start;
	    if ((*named_constraintp)->next==NULL) {
	      named_constraintp=NULL;
	      named_constraint_start=NULL;
	    } else {
	      named_constraintp=&(*named_constraintp)->next;
	      named_constraint_start=*named_constraintp;
	    }
	    free(p->s);
	    free(p);
	  } else {
	    free(p->next->s);
	    p->next=(*named_constraintp)->next;
	    free(*named_constraintp);
	    if (p->next==NULL) named_constraintp=NULL;
	    else named_constraintp=&(p->next);
	  }
	  s--;
	  free(s);
	  break;
	}
	s--;
	free(s);
	p=*named_constraintp;
	if ((*named_constraintp)->next==NULL) named_constraintp=NULL;
	else named_constraintp=&(*named_constraintp)->next;
      }
    }
    free(t);
  }
}

void Constr_Name_Set(GtkCellRendererText *cell, const gchar *path_string,
	     const gchar *new_text, gpointer data)
{
  GtkTreeSelection *sel;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkTextIter start,end;
  char *t,*s,*n;
  int type,a,b,doit;
  constrstr *p;

  named_constraintp=&named_constraint_start;
  while (*named_constraintp!=NULL) {
    s=strdup((*named_constraintp)->s);
    t=s+1;
    for (a=0;a<(int)strlen(t);a++) {
      if (t[a]==127) {
	t[a]='\0';break;
      }
    }
    if (strcmp(t,new_text)==0) {
      free(s);
      return;
    }
    free(s);
    if ((*named_constraintp)->next==NULL) named_constraintp=NULL;
    else named_constraintp=&(*named_constraintp)->next;
  }
  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(og.treeview));
  if (gtk_tree_selection_get_selected (sel, &model, &iter)) {
    gtk_tree_model_get(model, &iter,0,&t,1,&s,2,&type,-1);
    if (type) {
      /* file */
      free(s);
      free(t);
      gtk_text_buffer_get_bounds(og.showtbuff,&start,&end);
      s=gtk_text_buffer_get_text(og.showtbuff,&start,&end,FALSE);
      gtk_list_store_set(GTK_LIST_STORE(model),
			&iter,0,new_text,1,s,2,0,3,TRUE,-1);
      n=(char *)calloc(strlen(new_text)+strlen(s)+3,sizeof(char));
      n[0]=127;
      strcpy(&n[1],new_text);
      n[strlen(new_text)+1]=127;
      strcpy(&n[strlen(new_text)+2],s);
      for (a=strlen(new_text)+2;a<(int)strlen(n);a++) {
	if (n[a]=='\n') n[a]=127;
      }
      named_constraintp=&named_constraint_start;
      if (named_constraint_start==NULL) {
	(*named_constraintp)=(constrstr *)malloc(sizeof(constrstr));
	(*named_constraintp)->s=n;
	(*named_constraintp)->next=NULL;
      } else {
	while ((*named_constraintp)!=NULL) {
	  named_constraintp=&(*named_constraintp)->next;
	}
	(*named_constraintp)=(constrstr *)malloc(sizeof(constrstr));
	(*named_constraintp)->s=n;
	(*named_constraintp)->next=NULL;
      }
      /* free filename */
      doit=0;
      for (a=og.fhstart;a!=og.fhend;a=(a+1)%CH_Max) {
	if (strcmp(s,og.fnamehist[a])==0) {
	  free(og.fnamehist[a]);
	  doit=1;
	}
	if (doit) {
	  b=(a+1)%CH_Max;
	  og.fnamehist[a]=og.fnamehist[b];
	}
      }
      og.fhend=(og.fhend+CH_Max-1)%CH_Max;
      free(s);
    } else {
      /* named */
      gtk_list_store_set(GTK_LIST_STORE(model),
			&iter,0,new_text,1,s,2,0,3,TRUE,-1);
      named_constraintp=&named_constraint_start;
      p=NULL;
      while (named_constraintp!=NULL) {
	s=strdup((*named_constraintp)->s);
	s++;
	for (a=0;a<(int)strlen(s);a++) {
	  if (s[a]==127) {
	    s[a]='\0';
	    break;
	  }
	}
	if (strcmp(s,t)==0) {
	  n=(char *)calloc(strlen((*named_constraintp)->s)-strlen(s)+
	      strlen(new_text),sizeof(char));
	  n[0]=127;
	  strcpy(&n[1],new_text);
	  n[strlen(new_text)+1]=127;
	  strcpy(&n[strlen(new_text)+2],&s[a+1]);
	  free((*named_constraintp)->s);
	  (*named_constraintp)->s=n;
	  s--;
	  free(s);
	  break;
	}
	s--;
	free(s);
	p=*named_constraintp;
	if ((*named_constraintp)->next==NULL) named_constraintp=NULL;
	else named_constraintp=&(*named_constraintp)->next;
      }
      free(t);
    }
  }
}

void Load_Prev(GtkWidget *widget, gpointer data)
{
  Tree_Selection(gtk_tree_view_get_selection(GTK_TREE_VIEW(og.treeview)),data);
}

void Treeview_For_Constraint(int mode)
{
  GtkStyle  *style;		/* style for pixmap background */
  GdkPixmap *pixmap;		/* pixmap for pixmap buttons */
  GdkBitmap *mask;		/* pixmap mask */
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *sw;
  GtkWidget *button;
  GtkWidget *scroll;
  GtkWidget *text;
  GtkWidget *vpaned;
  GtkWidget *frame1;
  GtkWidget *frame2;
  GtkListStore *model;
  GtkTreeIter iter;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *col;
  GtkTreeSelection *selection;
#define SSIZE 20
  char s[SSIZE+1],*t;
  int a,b;

  /* create window, etc */
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window),"Insert/Load Previous Constraint");
  gtk_widget_realize(window);
  style= gtk_widget_get_style(window);
  pixmap = gdk_pixmap_create_from_xpm_d (window->window, &mask,
		       &style->bg[GTK_STATE_NORMAL],XPMicon);
  gdk_window_set_icon(window->window,NULL,pixmap,mask);
  gtk_container_set_border_width (GTK_CONTAINER (window), 5);
  g_signal_connect (window, "destroy",
		    G_CALLBACK (gtk_widget_destroy), &window);
  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  vpaned=gtk_vpaned_new();
  gtk_box_pack_start (GTK_BOX (vbox), vpaned, TRUE, TRUE, 0);
  frame1=gtk_frame_new(NULL);
  frame2=gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME (frame1), GTK_SHADOW_IN);
  gtk_frame_set_shadow_type(GTK_FRAME (frame2), GTK_SHADOW_IN);
  gtk_paned_pack1(GTK_PANED (vpaned), frame1, TRUE, TRUE);
  gtk_paned_pack2(GTK_PANED (vpaned), frame2, TRUE, TRUE);

  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_set_size_request(sw,100,200);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
				       GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (frame1), sw);

  model = gtk_list_store_new (4,G_TYPE_STRING,G_TYPE_STRING,
					G_TYPE_INT,G_TYPE_BOOLEAN);
  if (mode) {
    for (a=og.hstart;a!=og.hend;a=(a+1)%CH_Max) {
      strncpy(s,og.history[a],SSIZE);
      s[SSIZE]='\0';
      for (b=0;b<(int)strlen(s);b++) {
	if (s[b]=='\n') {
	  s[b]='\0';break;
	}
      }
      gtk_list_store_append(model, &iter);
      gtk_list_store_set(model, &iter,0,s,1,og.history[a],2,0,3,TRUE,-1);
    }
  } else {
    if (named_constraint_start!=NULL) {
      named_constraintp=&named_constraint_start;
      while (named_constraintp!=NULL) {
	t=strdup((*named_constraintp)->s);
	t++;
	for (a=0;a<(int)strlen(t);a++) {
	  if (t[a]==127) {
	    t[a++]='\0';
	    for (b=a;b<(int)strlen(&t[a]);b++) {
	      if (t[b]==127) t[b]='\n';
	    }
	    break;
	  }
	}
	gtk_list_store_append(model, &iter);
	gtk_list_store_set(model, &iter,0,t,1,&t[a],2,0,3,TRUE,-1);
	t--;
	free(t);
	if ((*named_constraintp)->next==NULL) named_constraintp=NULL;
	else named_constraintp=&(*named_constraintp)->next;
      }
    }
    for (a=og.fhstart;a!=og.fhend;a=(a+1)%CH_Max) {
      t=my_basename(og.fnamehist[a]);
      if (strlen(t)>SSIZE) t+=(strlen(t)-SSIZE);
      gtk_list_store_append(model, &iter);
      gtk_list_store_set(model, &iter,0,t,1,og.fnamehist[a],2,800,3,TRUE,-1);
    }
  }
  
  og.treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(model));
  g_object_unref (model);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (og.treeview), TRUE);
  selection=gtk_tree_view_get_selection(GTK_TREE_VIEW (og.treeview));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
  g_signal_connect(selection,"changed",G_CALLBACK(Tree_Selection),
      							(gpointer)mode);
  col = gtk_tree_view_column_new();
  gtk_tree_view_column_set_title(col,"Filename/Name");
  gtk_tree_view_append_column(GTK_TREE_VIEW(og.treeview), col);
  renderer = gtk_cell_renderer_text_new();
  g_signal_connect (renderer, "edited",G_CALLBACK(Constr_Name_Set),NULL);
  gtk_tree_view_column_pack_start(col, renderer, TRUE);
  gtk_tree_view_column_add_attribute(col,renderer,"text",0);
  gtk_tree_view_column_add_attribute(col,renderer,"editable",3);
  gtk_tree_view_column_add_attribute(col,renderer,"weight",2);
  gtk_container_add (GTK_CONTAINER (sw), og.treeview);

  hbox = gtk_hbox_new (TRUE, 4);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  button = gtk_button_new_with_label ("Load");
  g_signal_connect (button, "clicked", G_CALLBACK (Load_Prev),
						(gpointer)(2+mode));
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
	  GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) window);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  button = gtk_button_new_with_label ("Insert");
  g_signal_connect (button, "clicked", G_CALLBACK (Load_Prev),
						(gpointer)(4+mode));
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
	  GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) window);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  button = gtk_button_new_with_label ("Cancel");
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
	  GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) window);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  button = gtk_button_new_with_label ("Remove");
  g_signal_connect (button, "clicked",
		    G_CALLBACK (Remove_File_From_List), NULL);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  button = gtk_label_new("Rename if you want to preserve");
  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  button = gtk_label_new("Remove is only for list. File remains untouched.");
  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);

  scroll=gtk_scrolled_window_new(NULL,NULL);
  gtk_widget_set_size_request(scroll,100,100);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scroll),
						GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scroll),
			GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

  gtk_container_add (GTK_CONTAINER (frame2), scroll);
  og.showtbuff=gtk_text_buffer_new(NULL);
  text=gtk_text_view_new_with_buffer(GTK_TEXT_BUFFER(og.showtbuff));
  gtk_text_view_set_editable(GTK_TEXT_VIEW(text), FALSE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(text),GTK_WRAP_WORD);
  gtk_container_add(GTK_CONTAINER(scroll),text);

  gtk_window_set_default_size (GTK_WINDOW (window), 320, 200);
  gtk_widget_show_all(window);
}

void Filesel_For_Constraint(const char *s,void (*func) (GtkWidget *,gpointer))
{
  GtkWidget *filesel;

  filesel=gtk_file_selection_new("Load Constraints");
  gtk_signal_connect (GTK_OBJECT
	  (GTK_FILE_SELECTION(filesel)->ok_button), "clicked",
	  GTK_SIGNAL_FUNC (func), (gpointer)filesel);
  gtk_signal_connect_object (GTK_OBJECT
	  (GTK_FILE_SELECTION(filesel)->ok_button), "clicked",
	  GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) filesel);
  gtk_signal_connect_object (GTK_OBJECT
	  (GTK_FILE_SELECTION(filesel)->cancel_button), "clicked",
	  GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) filesel);
  gtk_widget_show(filesel);
}

/* constraint file functions */
void Constraint_File(GtkWidget *widget, gpointer data)
{
  int a,nok,p;
  GtkTextIter start,end;
  char *s,*t;

  a=(int)data;
  switch (a) {
    case 0:
      gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
      gtk_text_buffer_delete(og.tbuff, &start,&end);
      break;
    case 1:
      Filesel_For_Constraint("Load Constraint",Load_Constraint);
      break;
    case 2:
      Filesel_For_Constraint("Insert Constraint",Insert_Constraint);
      break;
    case 3:
      if ((og.fhstart==og.fhend) && (named_constraint_start==NULL)) break;
      Treeview_For_Constraint(0);
      break;
    case 4:
      if (og.hstart==og.hend) break;
      Treeview_For_Constraint(1);
      break;
    case 5:
      Filesel_For_Constraint("Save Constraint",Save_Constr_As);
      break;
    case 6:
      gtk_text_buffer_get_bounds(og.tbuff, &start,&end);
      s=strdup(gtk_text_buffer_get_text(og.tbuff,&start,&end,FALSE));
      for (a=0;a<(int)strlen(s);a++) {
	if (s[a]=='\n') s[a]=127;
      }
      t=(char *)calloc(strlen(s)+strlen("save")+20,sizeof(char));
      sprintf(t,"%cname%c",127,127);
      a=strlen(t)-1;p=1;
      do {
	nok=0;
	named_constraintp=&named_constraint_start;
	while ((*named_constraintp)!=NULL) {
	  if (strncmp(t,(*named_constraintp)->s,a)==0) nok=1;
	  named_constraintp=&(*named_constraintp)->next;
	}
	if (nok) {
	  sprintf(t,"%csave%d%c",127,p++,127);
	  a=strlen(t)-1;
	}
      } while (nok);
      strcat(t,s);
      (*named_constraintp)=(constrstr *)malloc(sizeof(constrstr));
      (*named_constraintp)->s=t;
      (*named_constraintp)->next=NULL;
      free(s);
      Treeview_For_Constraint(0);
      break;
  }
}

void reset_og(void)
{
  og.numchange=1.0;
  RPN_deg=1;
}

void CreateObjectFrame(GtkWidget *frame)
{
  GtkWidget *vbox;		/* vbox */
  GtkWidget **hbox;		/* hboxes */
  GtkWidget **button;		/* Buttons */
  GtkWidget **oonoff;		/* chkbutton for objects */
  GtkWidget *off;		/* switch frame off button */
  GtkWidget *update;		/* Update button */
  GtkWidget *cbutton[CNN];	/* costraints */
  constrstr *p;
  char *s;
  int i;

  if (bg.obj_num==0) return;
  bg.cgnum=0;
  og.istab=0;
  og.hstart=og.hindex=og.hend=0;
  og.fhstart=og.fhend=0;
  og.variablehelpisup=0;
  og.functionhelpisup=0;
  og.setnumoriginal=0;
  /* globarl variables for RPN */
  sprintf(bg.rpnglobstr[0],"gl.x");
  sprintf(bg.rpnglobstr[1],"gl.y");
  sprintf(bg.rpnglobstr[2],"gl.z");
  sprintf(bg.rpnglobstr[3],"gl.phix");
  sprintf(bg.rpnglobstr[4],"gl.phiy");
  sprintf(bg.rpnglobstr[5],"gl.phiz");
  bg.rpnglobpoint[0]=&gl.x;
  bg.rpnglobpoint[1]=&gl.y;
  bg.rpnglobpoint[2]=&gl.z;
  bg.rpnglobpoint[3]=&gl.phix;
  bg.rpnglobpoint[4]=&gl.phiy;
  bg.rpnglobpoint[5]=&gl.phiz;

  /* restore rc */
  mparam_cstrp=&mparam_cstr_start;
  named_constraint_start=NULL;
  named_constraintp=&named_constraint_start;
  if (mparam_cstr_start!=NULL) {
    while (mparam_cstrp!=NULL) {
      s=(*mparam_cstrp)->s;
      if (s[0]==127) {
	/* name + constraint */
	*named_constraintp=(constrstr *) malloc(sizeof(constrstr));
	(*named_constraintp)->s=s;
	(*named_constraintp)->next=NULL;
	named_constraintp=&((*named_constraintp)->next);
      } else {
	/* filename */
	if (og.fhend<CH_Max) og.fnamehist[og.fhend++]=s;
      }
      p=(*mparam_cstrp);
      if ((*mparam_cstrp)->next==NULL) mparam_cstrp=NULL;
      else mparam_cstrp=&(*mparam_cstrp)->next;
    }
    mparam_cstrp=&mparam_cstr_start;
    while (mparam_cstrp!=NULL) {
      p=(*mparam_cstrp);
      if ((*mparam_cstrp)->next==NULL) mparam_cstrp=NULL;
      else mparam_cstrp=&(*mparam_cstrp)->next;
      free(p);
    }
  }
  cvg.var_num=body_alloc*3;
  cvg.var=(vect *)calloc(sizeof(vect),cvg.var_num*3);
  cvg.vc=(int *)calloc(sizeof(int),cvg.var_num);
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(frame),vbox);
  hbox=(GtkWidget **) calloc(bg.obj_num/3+1,sizeof(GtkWidget *));
  button=(GtkWidget **) calloc(bg.obj_num,sizeof(GtkWidget *));
  oonoff=(GtkWidget **) calloc(bg.obj_num,sizeof(GtkWidget *));
  for (i=0;i<bg.obj_num/3+1;i++) {
    hbox[i] = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox),hbox[i],FALSE,FALSE,0);
    gtk_widget_show(hbox[i]);
  }
  for (i=0;i<bg.obj_num;i++) {
    s=(char *)calloc(strlen(obj_name[bg.param[i].type])+
		ilog((double)i+1.0)+10,sizeof(char));
    sprintf(s,"%s(%d)",obj_name[bg.param[i].type],i+1);
    oonoff[i]=gtk_check_button_new_with_label(s);
    free(s);
    gtk_box_pack_start(GTK_BOX(hbox[i/3]),oonoff[i],FALSE,FALSE,5);
    if (anim.init_show) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(oonoff[i]), TRUE);
    }
    gtk_signal_connect(GTK_OBJECT (oonoff[i]),"clicked",
            GTK_SIGNAL_FUNC(Show_Onoff),(gpointer)i);
    gtk_widget_show(oonoff[i]);
  }
  off=gtk_check_button_new_with_label("Off");
  gtk_box_pack_end(GTK_BOX(hbox[(bg.obj_num-1)/3]),off,FALSE,FALSE,0);
  gtk_widget_show(off);

  for (i=0;i<2;i++) {
    og.chbox[i]=gtk_hbox_new(FALSE,0);
    gtk_box_pack_start(GTK_BOX(vbox),og.chbox[i],FALSE,FALSE,0);
  }
  for (i=0;i<CNN;i++) {
    cbutton[i]=gtk_button_new_with_label(constr_name[i]);
    gtk_box_pack_start(GTK_BOX(og.chbox[i/(CNN/2+1)]),cbutton[i],
							FALSE,FALSE,0);
    gtk_signal_connect(GTK_OBJECT (cbutton[i]),"clicked",
            GTK_SIGNAL_FUNC(Constraint_Insert_Text),
	    (gpointer)constr_name[i]);
    gtk_widget_show(cbutton[i]);
  }
  for (i=0;i<2;i++) {
    gtk_widget_show(og.chbox[i]);
  }

  og.bhbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),og.bhbox,FALSE,FALSE,3);
  update=gtk_button_new_with_label("Update");
  gtk_box_pack_start(GTK_BOX(og.bhbox),update,FALSE,FALSE,3);
  gtk_widget_show(update);
  og.valuechangetext=gtk_label_new("Value change:");
  gtk_box_pack_start(GTK_BOX(og.bhbox),og.valuechangetext,FALSE,FALSE,0);
  og.valuechangeval=gtk_entry_new_with_max_length(50);
  gtk_widget_set_usize(og.valuechangeval,70,0);
  gtk_box_pack_start(GTK_BOX(og.bhbox),og.valuechangeval,FALSE,FALSE,0);
  og.status=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(og.bhbox),og.status,FALSE,FALSE,3);
  gtk_widget_show(og.status);
  gtk_widget_show(og.bhbox);

  og.scroll=gtk_scrolled_window_new(NULL,NULL);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(og.scroll),
						GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(og.scroll),
			GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

  gtk_box_pack_end(GTK_BOX(vbox),og.scroll,TRUE,TRUE,0);
  og.tbuff=gtk_text_buffer_new(NULL);
  og.text=gtk_text_view_new_with_buffer(GTK_TEXT_BUFFER(og.tbuff));
  gtk_text_view_set_editable(GTK_TEXT_VIEW(og.text), TRUE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(og.text),GTK_WRAP_WORD);
  gtk_container_add(GTK_CONTAINER(og.scroll),og.text);
  gtk_text_buffer_create_tag (og.tbuff, "bold",
				"weight", PANGO_WEIGHT_BOLD, NULL);  


  gtk_widget_show(og.scroll);
  gtk_widget_show(og.text);
  gtk_widget_show(vbox);

  og.constr_signalh=g_signal_connect(G_OBJECT (og.tbuff), "changed",
		G_CALLBACK (Constraint_Insert), NULL);
  g_signal_connect(G_OBJECT (og.tbuff), "mark_set",
		G_CALLBACK (Constraint_Cursor), NULL);

  gtk_signal_connect (GTK_OBJECT (update), "clicked",
		GTK_SIGNAL_FUNC (Update_Constraint), 0 );
  gtk_signal_connect(GTK_OBJECT (off), "toggled",
		GTK_SIGNAL_FUNC (FrameOff), (gpointer)frame);
}

