/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2005/03/01 13:59:36 $ $Revision: 1.30 $ */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "global.h"
#include "io.h"
#include "anim.h"
#include "light.h"
#include "rpn.h"
#include "body.h"

#ifdef _WIN32
/*'conversion' conversion from 'type1' to 'type2', possible loss of data*/
#pragma warning(disable:4244)
#endif

static double *constr_global_pointer[6]={ &gl.x, &gl.y, &gl.z,
		&gl.phix, &gl.phiy, &gl.phix };

int ilog(double);
void perr(char *, ...);
void io_setfiletype(int);
int io_detfiletype(void);
int move_bp(char *,int);
double RPN_Evaluate(RPN *,double *,double **,double *,vect *,int *,int,int);
void Plot_Font(constrglobal *,int);

/* These settings are valid for all objects */
void Set_General(void)
{
  glEnable(GL_COLOR_MATERIAL);
  glColor4f(lb.color[0],lb.color[1],lb.color[2],lb.color[3]);
  if (lb.tex) {
    glBindTexture(GL_TEXTURE_2D,gl.texture.texture[lb.tex-1]);
    glEnable(GL_TEXTURE_2D);
  } else {
    glDisable(GL_TEXTURE_2D);
  }

  if (lb.blend) {
    glEnable(GL_BLEND);
  } else glDisable(GL_BLEND);

  glLoadIdentity();
  glTranslatef(gl.x,gl.y,gl.z);
  glRotatef(gl.phix,1.0,0.0,0.0);
  glRotatef(gl.phiz,0.0,0.0,1.0);
  glRotatef(gl.phiy,0.0,1.0,0.0);
  glRotatef(-90.0,1.0,0.0,0.0);
  glTranslatef(gl.f*lb.x, gl.f*lb.y,gl.f*lb.z);
  glRotatef(lb.phi*RP,lb.rx,lb.ry,lb.rz);
}

void Set_Dir(vect *vrv,double *rx,double *ry,double *rz,double *phi)
{
  double srv;

  srv=sqrt(vrv->v[0]*vrv->v[0]+vrv->v[1]*vrv->v[1]+vrv->v[2]*vrv->v[2]);
  if (srv>0.0) {
    vrv->v[0]/=srv;
    vrv->v[1]/=srv;
  }
  *rx=vrv->v[1];
  *ry=-vrv->v[0];
  *rz=0.0;
  if ((fabs(*rx)+fabs(*ry))==0.0) {
    *phi=0.0;
    *rx=1.0;
  } else {
    *phi=-asin(sqrt(vrv->v[0]*vrv->v[0]+vrv->v[1]*vrv->v[1]));
  }
  if (vrv->v[2]<0.0) *phi=M_PI-(*phi);
}

int DetArray(int a,int b,int n,int *Ne,int *Np,int *atype,int *ntype)
{
  int N,c;
  double srv;
  
  N=n-1;
  if (cg[b].p[N]!=NULL) {
    RPN_Evaluate(cg[b].p[N],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
    *ntype=*atype=(int)srv;
    if (*atype<0) {
      *ntype=-1;
      *atype*=-1;
    }
  } else {
    *ntype=*atype=-1;
  }
  N--;
  if ((cg[b].p[N]==NULL) || (cg[b].arrayvar[N]==0)) *Ne=0;
  else *Ne=1;
  *Np=0;
  while (N>0) {
    N--;
    if ((cg[b].p[N]!=NULL) && (cg[b].arrayvar[N])) *Np=1;
  }
  /* not of this type */
  if ((*atype>0) && (*atype!=body[a].obj+1)) return(2);
  N=n-2; /* expr */
  if (!(*Ne)) {
    RPN_Evaluate(cg[b].p[N],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
    if (srv<=0.0) return(1);
  }
  if ((*Ne) && (!(*Np))) {
    for (c=0;c<bg.body_num;c++) {
      if ((*ntype<0) || (*ntype==body[c].obj+1)) {
        RPN_Evaluate(cg[b].p[N],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
      }
    }
    RPN_Evaluate(cg[b].p[N],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
    if (srv<=0.0) return(1);
  }
  return(0);
}

void SetTrans(void)
{
  int a,b,type;
  double srv;

  a=0;
  for (b=0;b<bg.cgnum;b++) {
    if (cg[b].type==10) {
      if (cg[b].p[0]==NULL) continue;
      a=1;
    }
  }
  if (!a) return;
  for (a=0;a<bg.body_num;a++) {
    body[a].blend=body[a].appearance&BA_Blend;
  }
  for (b=0;b<bg.cgnum;b++) {
    if (cg[b].type==10) {
      if (cg[b].p[0]==NULL) continue;
      type=-1;
      if (cg[b].p[2]!=NULL) {
        RPN_Evaluate(cg[b].p[2],body[0].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,0);
        type=(int)srv;
      }
      for (a=0;a<bg.body_num;a++) {
	if ((body[a].obj+1!=type) && (type!=-1)) continue;
	srv=1.0;
	if (cg[b].p[1]!=NULL) {
	  RPN_Evaluate(cg[b].p[1],body[a].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
	}
	if (srv>0.0) {
	  RPN_Evaluate(cg[b].p[0],body[a].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
	  body[a].blend=(srv>0.0);
	}
      }
    }
  }
}

void SetGlobalVar(int i,vect *vrv,int sv)
{
  if (i<6) {
    (*constr_global_pointer[i])=vrv->v[0];
  } else if (sv) {
    if (i==6) {
      gl.x=vrv->v[0];
      gl.y=vrv->v[1];
      gl.z=vrv->v[2];
    } else {
      gl.phix=vrv->v[0];
      gl.phiy=vrv->v[1];
      gl.phiz=vrv->v[2];
    }
  }
}

void SetVar(void)
{
  int a,b,n,atype,ntype,cycle,vari,sv,change;
  double srv;
  vect vrv;

  if ((anim.objectread==0) && (animation.framerepeat<=1)) return;
  anim.objectread=0;
  for (b=0;b<bg.cgnum;b++) {
    if (cg[b].type==9) {
      /* Var */
      if ((cg[b].p[0]==NULL) || (cg[b].p[1]==NULL)) continue;
      if (!cg[b].arrayvar[0]) {
	if (cg[b].p[0]->type==RpnT_GlobalVar) {
	  vari=-1-cg[b].p[0]->ival;
	} else {
	  RPN_Evaluate(cg[b].p[0],body[0].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,0);
	  vari=(int)srv;
	  if ((vari<0) || (vari>=cvg.var_num)) continue;
	}
      }
      if (cg[b].p[3]!=NULL) {
        RPN_Evaluate(cg[b].p[3],body[0].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,0);
        ntype=atype=(int)srv;
        if (atype<0) {
          ntype=-1;
          atype*=-1;
        }
      } else {
        ntype=atype=-1;
      }
      if (cg[b].p[2]==NULL) cycle=0;
      else if (cg[b].arrayvar[2]==0) cycle=1;
      else cycle=2;
      if (cg[b].arrayvar[1]!=0) cycle=2;
      if (vari<0) cycle=min(1,cycle);
      if (cycle==0) {
	RPN_Evaluate(cg[b].p[1],body[0].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,0,0);
	if (vari>=0) {
	  cvg.var[vari].v[0]=vrv.v[0];
	  cvg.var[vari].v[1]=vrv.v[1];
	  cvg.var[vari].v[2]=vrv.v[2];
	  cvg.vc[vari]=sv;
	} else {
	  SetGlobalVar(-1-vari,&vrv,sv);
	}
      } else if (cycle==1) {
	if (vari>=0) {
	  change=0;
	  for (a=0;a<bg.body_num;a++) {
	    if ((atype>0) && (atype!=body[a].obj+1)) continue;
	    RPN_Evaluate(cg[b].p[0],body[a].params,bg.rpnglobpoint,
				&srv,NULL,NULL,4,a);
	    if (vari!=(int)srv) change=1;
	    vari=(int)srv;
	    if ((vari>=0) && (vari<cvg.var_num)) {
	      RPN_Evaluate(cg[b].p[2],body[a].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
	      if (srv>0.0) {
		RPN_Evaluate(cg[b].p[1],body[a].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,3,a);
		cvg.var[vari].v[0]=vrv.v[0];
		cvg.var[vari].v[1]=vrv.v[1];
		cvg.var[vari].v[2]=vrv.v[2];
		cvg.vc[vari]=sv;
	      }
	    }
	  }
	  if (!change) {
	    RPN_Evaluate(cg[b].p[1],body[0].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,4,0);
	    cvg.var[vari].v[0]=vrv.v[0];
	    cvg.var[vari].v[1]=vrv.v[1];
	    cvg.var[vari].v[2]=vrv.v[2];
	    cvg.vc[vari]=sv;
	  }
	} else {
	  change=0;
	  for (a=0;a<bg.body_num;a++) {
	    if ((atype>0) && (atype!=body[a].obj+1)) continue;
	    RPN_Evaluate(cg[b].p[2],body[a].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
	    if (srv>0.0) {
	      RPN_Evaluate(cg[b].p[1],body[a].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,3,a);
	      SetGlobalVar(-1-vari,&vrv,sv);
	    }
	  }
	  if (!change) {
	    RPN_Evaluate(cg[b].p[1],body[0].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,4,0);
	    SetGlobalVar(-1-vari,&vrv,sv);
	  }
	}
      } else { /* cycle===2 */
        for (a=0;a<bg.body_num;a++) {
          if ((atype>0) && (atype!=body[a].obj+1)) continue;
          for (n=0;n<bg.body_num;n++) {
            if ((ntype>0) && (ntype!=body[n].obj+1)) continue;
            RPN_Evaluate(cg[b].p[2],body[a].params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,n);
            if (srv>0.0) {
              RPN_Evaluate(cg[b].p[1],body[a].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,3,n);
            }
          }
          if (cg[b].arrayvar[0]!=0) {
            RPN_Evaluate(cg[b].p[0],body[a].params,bg.rpnglobpoint,
				&srv,NULL,NULL,4,a);
            vari=(int)srv;
            if ((vari>=0) && (vari<cvg.var_num)) {
              RPN_Evaluate(cg[b].p[1],body[a].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,4,a);
              cvg.var[vari].v[0]=vrv.v[0];
              cvg.var[vari].v[1]=vrv.v[1];
              cvg.var[vari].v[2]=vrv.v[2];
              cvg.vc[vari]=sv;
            }
          }
        }
        if (cg[b].arrayvar[0]==0) {
          RPN_Evaluate(cg[b].p[1],body[a].params,bg.rpnglobpoint,
				NULL,&vrv,&sv,4,0);
          cvg.var[vari].v[0]=vrv.v[0];
          cvg.var[vari].v[1]=vrv.v[1];
          cvg.var[vari].v[2]=vrv.v[2];
          cvg.vc[vari]=sv;
        }
      }
    }
  }
}

void CheckFont(int a)
{
  int b;

  if (a) {
    for (b=0;b<bg.cgnum;b++) {
      if (cg[b].type==-11) {
	Plot_Font(&cg[b],a);
      }
    }
  } else {
    for (b=0;b<bg.cgnum;b++) {
      if (cg[b].type==11) {
	Plot_Font(&cg[b],a);
      }
    }
  }
}

int Constraint(int a,int b,int mode)
{
  int c,d,atype,ntype,Ne,Np;
  int sv;
  double srv,srv1;
  double px,py,pz,cpy,spy;
  vect vrv;

  switch (cg[b].type) {
    case 0:     /* Arrow */
      if (DetArray(a,b,5,&Ne,&Np,&atype,&ntype)) return(0);
      if (Np) {
        for (c=0;c<bg.body_num;c++) {
          if ((ntype<0) || (ntype==body[c].obj+1)) {
            if (Ne) RPN_Evaluate(cg[b].p[3],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,2,c);
            else srv=1.0;
            if (srv>0) {
	      RPN_Evaluate(cg[b].p[2],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
	      RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
              RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
            }
          }
        }
      }
      RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,&vrv,&sv,0,a);
      if (sv) {
        lb.type=BT_Arrow;
        Set_Dir(&vrv,&lb.rx,&lb.ry,&lb.rz,&lb.phi);
        if (cg[b].p[1]!=NULL) {
          RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
          lb.sx=srv;
        }
        if (cg[b].p[2]!=NULL) {
          RPN_Evaluate(cg[b].p[2],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
          lb.sz=srv;
        } else {
          lb.sz=lb.sx*0.01;
        }
      }
      break;
    case 1:     /* Color */
      if (DetArray(a,b,6,&Ne,&Np,&atype,&ntype)) return(0);
      if (Np) {
        for (c=0;c<bg.body_num;c++) {
          if ((ntype<0) || (ntype==body[c].obj+1)) {
            if (Ne) RPN_Evaluate(cg[b].p[4],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,2,c);
            else srv=1.0;
            if (srv>0) {
              RPN_Evaluate(cg[b].p[3],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
              RPN_Evaluate(cg[b].p[2],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
              RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
              RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
            }
          }
        }
      }
      if (cg[b].p[3]!=NULL) {
        RPN_Evaluate(cg[b].p[3],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,mode,0);
        lb.color[3]=srv;
      }
      RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				&srv,&vrv,&sv,mode,0);
      if (sv) {
	lb.color[0]=vrv.v[0];
	lb.color[1]=vrv.v[1];
	lb.color[2]=vrv.v[2];
      } else {
	lb.color[0]=srv;
	if (cg[b].p[2]!=NULL) {
	  RPN_Evaluate(cg[b].p[2],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,mode,0);
	  lb.color[2]=srv;
	}
	if (cg[b].p[1]!=NULL) {
	  RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,mode,0);
	  lb.color[1]=srv;
	}
      }
      break;
    case 2:     /* Cut */
      if (DetArray(a,b,4,&Ne,&Np,&atype,&ntype)) return(0);
      if (Np) {
        for (c=0;c<bg.body_num;c++) {
          if ((ntype<0) || (ntype==body[c].obj+1)) {
            if (Ne) RPN_Evaluate(cg[b].p[2],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,2,c);
            else srv=1.0;
            if (srv>0) {
              RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
              RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
            }
          }
        }
      }
      RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,mode,0);
      RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv1,NULL,NULL,mode,0);
      px=gl.phix/RP;
      py=gl.phiy/RP;
      pz=gl.phiz/RP;
      cpy=cos(py);
      spy=sin(py);
      if (fabs(-cos(px)*(lb.x*spy+lb.y*cpy)+
                sin(px)*( (lb.x*cpy-lb.y*spy)*sin(pz)+
                lb.z*cos(pz))+srv) > fabs(srv1)) return(1);
      break;
    case 3:     /* Dir */
      if (DetArray(a,b,3,&Ne,&Np,&atype,&ntype)) return(0);
      if (Np) {
        for (c=0;c<bg.body_num;c++) {
          if ((ntype<0) || (ntype==body[c].obj+1)) {
            if (Ne) RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,2,c);
            else srv=1.0;
            if (srv>0) {
              RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
            }
          }
        }
      }
      RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,&vrv,&sv,mode,0);
      if (sv) {
        Set_Dir(&vrv,&lb.rx,&lb.ry,&lb.rz,&lb.phi);
        lb.params[CEF_SUBPARNUM*3]=lb.rx;
        lb.params[CEF_SUBPARNUM*3+1]=lb.ry;
        lb.params[CEF_SUBPARNUM*3+2]=lb.rz;
        lb.params[CEF_SUBPARNUM*3+3]=lb.phi;
      }
      break;
    case 4:     /* New */
      /* do not recreate twice */
      if ((bg.new_num<=bg.body_num) || (a>=bg.body_num)) break;
      if ((cg[b].p[0]==NULL) || (cg[b].p[2]==NULL)) break;
      srv=2.0;
      if ((cg[b].p[1]==NULL) && ((a!=0) || (mode!=0))) break;
      Np=cg[b].arrayvar[1]+cg[b].arrayvar[2]+cg[b].arrayvar[3]+
						cg[b].arrayvar[4];
      Np=(Np ? bg.body_num : 1);
      for (c=0;c<Np;c++) {
	if (cg[b].p[1]!=NULL) 
	      RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,mode,c);
	if (srv>0.0) {
	  bg.new_num--;
	  RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,mode,c);
	  body[bg.new_num].type=(int)srv;
	  RPN_Evaluate(cg[b].p[2],lb.params,bg.rpnglobpoint,
				NULL,&vrv,NULL,mode,c);
	  body[bg.new_num].sx=vrv.v[0];
	  body[bg.new_num].sy=vrv.v[1];
	  body[bg.new_num].sz=vrv.v[2];
	  if (cg[b].p[3]!=NULL) {
	    RPN_Evaluate(cg[b].p[3],lb.params,bg.rpnglobpoint,
				NULL,&vrv,NULL,mode,c);
	    body[bg.new_num].x=vrv.v[0];
	    body[bg.new_num].y=vrv.v[1];
	    body[bg.new_num].z=vrv.v[2];
	  } else {
	    body[bg.new_num].x=0.0;
	    body[bg.new_num].y=0.0;
	    body[bg.new_num].z=0.0;
	  }
	  if (cg[b].p[4]!=NULL) {
	    RPN_Evaluate(cg[b].p[4],lb.params,bg.rpnglobpoint,
				NULL,&vrv,NULL,mode,c);
	    Set_Dir(&vrv,&body[bg.new_num].rx,&body[bg.new_num].ry,
			&body[bg.new_num].rz,&body[bg.new_num].phi);
	  } else {
	    body[bg.new_num].rx=1.0;
	    body[bg.new_num].ry=0.0;
	    body[bg.new_num].rz=0.0;
	    body[bg.new_num].phi=0.0;
	  }
	  for (d=bg.new_num+1;d<body_alloc;d++) {
	    if ((body[d].type==body[bg.new_num].type) &&
		(body[d].sx==body[bg.new_num].sx) &&
		(body[d].sy==body[bg.new_num].sy) &&
		(body[d].sz==body[bg.new_num].sz) &&
		(body[d].x==body[bg.new_num].x) &&
		(body[d].y==body[bg.new_num].y) &&
		(body[d].z==body[bg.new_num].z) &&
		(body[d].rx==body[bg.new_num].rx) &&
		(body[d].ry==body[bg.new_num].ry) &&
		(body[d].rz==body[bg.new_num].rz) &&
		(body[d].phi==body[bg.new_num].phi)) break;
	  }
	  if (d<body_alloc) {
	    bg.new_num++;
	  } else {
	    if (cg[b].p[5]!=NULL) {
	      RPN_Evaluate(cg[b].p[5],lb.params,bg.rpnglobpoint,
				NULL,&vrv,NULL,mode,c);
	      body[bg.new_num].color[0]=vrv.v[0];
	      body[bg.new_num].color[1]=vrv.v[1];
	      body[bg.new_num].color[2]=vrv.v[2];
	    } else {
	      body[bg.new_num].color[0]=1.0;
	      body[bg.new_num].color[1]=1.0;
	      body[bg.new_num].color[2]=1.0;
	    }
	    body[bg.new_num].color[3]=1.0;
	    body[bg.new_num].tex=0;
	    body[bg.new_num].blend=0;
	    body[bg.new_num].obj=bg.obj_num;
	  }
	}
      }
      break;
    case 5:     /* Position */
      if (DetArray(a,b,3,&Ne,&Np,&atype,&ntype)) return(0);
      if (Np) {
        for (c=0;c<bg.body_num;c++) {
          if ((ntype<0) || (ntype==body[c].obj+1)) {
            if (Ne) RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,2,c);
            else srv=1.0;
            if (srv>0) {
              RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
            }
          }
        }
      }
      RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,&vrv,&sv,mode,0);
      if (sv) {
        lb.params[CEF_SUBPARNUM*2]=lb.x=vrv.v[0];
        lb.params[CEF_SUBPARNUM*2+1]=lb.y=vrv.v[1];
        lb.params[CEF_SUBPARNUM*2+2]=lb.z=vrv.v[2];
      }
      break;
    case 6:     /* Texture */
      if (DetArray(a,b,3,&Ne,&Np,&atype,&ntype)) return(0);
      if (Np) {
        for (c=0;c<bg.body_num;c++) {
          if ((ntype<0) || (ntype==body[c].obj+1)) {
            if (Ne) RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,2,c);
            else srv=1.0;
            if (srv>0) {
              RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
            }
          }
        }
      }
      RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,0,a);
      lb.tex=(int)srv;
      break;
    case 7:     /* Show */
      if (DetArray(a,b,2,&Ne,&Np,&atype,&ntype)==1) return(1);
      break;
    case 8:     /* Size */
      if (DetArray(a,b,3,&Ne,&Np,&atype,&ntype)) return(0);
      if (Np) {
        for (c=0;c<bg.body_num;c++) {
          if ((ntype<0) || (ntype==body[c].obj+1)) {
            if (Ne) RPN_Evaluate(cg[b].p[1],lb.params,bg.rpnglobpoint,
				&srv,NULL,NULL,2,c);
            else srv=1.0;
            if (srv>0) {
              RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,NULL,NULL,2,c);
            }
          }
        }
      }
      RPN_Evaluate(cg[b].p[0],lb.params,bg.rpnglobpoint,
				NULL,&vrv,&sv,0,a);
      if (sv) {
        lb.params[CEF_SUBPARNUM]=lb.sx=vrv.v[0];
        lb.params[CEF_SUBPARNUM+1]=lb.sy=vrv.v[1];
        lb.params[CEF_SUBPARNUM+1]=lb.sz=vrv.v[2];
      } else {
        lb.params[CEF_SUBPARNUM]=lb.sx=vrv.v[0];
      }
      break;
  }
  return(0);
}

void Plot_Body(int a,int mode)
{
  int b,rv;
  double r,q;

  if ((a+anim.constrreread) && (body[a].obj!=bg.obj_num) &&
      (!bg.param[body[a].obj].show)) return;
  if ((mode) && (!body[a].blend)) return;
  if ((!mode) && (body[a].blend)) return;
  if ((!anim.donotdoublebody) && (anim.constrreread==1)) {
    if (!restb[a].show) return;
    lb.type=restb[a].type;
    lb.blend=restb[a].blend;
    lb.tex=restb[a].tex;
    lb.color[0]=restb[a].color[0];
    lb.color[1]=restb[a].color[1];
    lb.color[2]=restb[a].color[2];
    lb.color[3]=restb[a].color[3];
    lb.x=restb[a].x;
    lb.y=restb[a].y;
    lb.z=restb[a].z;
    lb.rx=restb[a].rx;
    lb.ry=restb[a].ry;
    lb.rz=restb[a].rz;
    lb.phi=restb[a].phi;
    lb.sx=restb[a].sx;
    lb.sy=restb[a].sy;
    lb.sz=restb[a].sz;
  } else {
    lb.type=body[a].type;
    lb.blend=body[a].blend;
    lb.tex=body[a].tex;
    lb.color[0]=body[a].color[0];
    lb.color[1]=body[a].color[1];
    lb.color[2]=body[a].color[2];
    lb.color[3]=body[a].color[3];
    lb.x=body[a].x;
    lb.y=body[a].y;
    lb.z=body[a].z;
    lb.rx=body[a].rx;
    lb.ry=body[a].ry;
    lb.rz=body[a].rz;
    lb.phi=body[a].phi;
    lb.sx=body[a].sx;
    lb.sy=body[a].sy;
    lb.sz=body[a].sz;
    for (b=0;b<CEF_PARNUM*CEF_SUBPARNUM;b++) {
      lb.params[b]=body[a].params[b];
    }
    if (a==0) {
      rv=bg.param[body[a].obj].show;
      for (b=0;b<bg.cgnum;b++) {
	if (Constraint(a,b,mode)) {
	  if (!anim.donotdoublebody) {
	    restb[a].type=lb.type;
	    restb[a].show=0;
	  }
	  rv=0;
	}
      }
      if (!rv) return;
    } else {
      for (b=0;b<bg.cgnum;b++) {
	if (Constraint(a,b,mode)) {
	  if (!anim.donotdoublebody) {
	    restb[a].type=lb.type;
	    restb[a].show=0;
	  }
	  return;
	}
      }
    }
    if (!anim.donotdoublebody) {
      restb[a].show=1;
      restb[a].type=lb.type;
      restb[a].blend=lb.blend;
      restb[a].tex=lb.tex;
      restb[a].color[0]=lb.color[0];
      restb[a].color[1]=lb.color[1];
      restb[a].color[2]=lb.color[2];
      restb[a].color[3]=lb.color[3];
      restb[a].x=lb.x;
      restb[a].y=lb.y;
      restb[a].z=lb.z;
      restb[a].rx=lb.rx;
      restb[a].ry=lb.ry;
      restb[a].rz=lb.rz;
      restb[a].phi=lb.phi;
      restb[a].sx=lb.sx;
      restb[a].sy=lb.sy;
      restb[a].sz=lb.sz;
    }
  }
  Set_General();
  r=lb.sx*gl.f*0.5;
  switch (lb.type) {
    case BT_Wall:
      q=lb.sy*0.5*gl.f;
      glBegin(GL_QUADS);
      glNormal3f(0.0,0.0,1.0);
      glTexCoord2f(0.0, 1.0);
      glVertex3f(-r, q, 0.0);
      glTexCoord2f(0.0, 0.0);
      glVertex3f(-r,-q, 0.0);
      glTexCoord2f(1.0, 0.0);
      glVertex3f( r,-q, 0.0);
      glTexCoord2f(1.0, 1.0);
      glVertex3f( r, q, 0.0);
      glEnd();
      break;
    case BT_Ellipsoid:
      glScalef(lb.sx*gl.f,lb.sy*gl.f,lb.sz*gl.f);
      r=0.5;
    case BT_DGBall:
    case BT_STBall:
    case BT_BallCC:
    case BT_Ball:
      gluSphere(gl.quad.q, r, gl.quad.sl,gl.quad.st);
      break;
    case BT_Force:
    case BT_DTEdge:
    case BT_Cylinder:
      gluCylinder(gl.quad.q,r,r,lb.sz*gl.f,gl.quad.sl,gl.quad.st);
      break;
    case BT_Disk:
      gluDisk(gl.quad.q, 0.0, r, gl.quad.sl,gl.quad.st);
      break;
    case BT_Arrow:
      q=lb.sz*gl.f;
      r*=2.0;
      gluCylinder(gl.quad.q,q,0.0,r,gl.quad.sl,gl.quad.st);
      glTranslatef(0.0,0.0,r*0.7);
      gluCylinder(gl.quad.q,2.0*q,0.0,r*0.3,gl.quad.sl,gl.quad.st);
      break;
    case BT_Cone:
      gluCylinder(gl.quad.q,r,lb.sy*gl.f*0.5,lb.sz*gl.f,gl.quad.sl,gl.quad.st);
      break;
    case BT_Triangle:
      r*=2.0;
      glBegin(GL_QUADS);
      glNormal3f(0.0,0.0,1.0);
      glTexCoord2f(0.0, 1.0);
      glVertex3f(0.0,0.0, 0.0);
      glTexCoord2f(0.0, 0.0);
      glVertex3f(0.0,0.0, 0.0);
      glTexCoord2f(1.0, 0.0);
      glVertex3f( r,0.0, 0.0);
      glTexCoord2f(1.0, 1.0);
      glVertex3f(lb.sy*gl.f,lb.sz*gl.f, 0.0);
      glEnd();
      break;
    case BT_Triangle2:
      r*=2.0;
      glBegin(GL_QUADS);
      glNormal3f(0.0,0.0,1.0);
      glTexCoord2f(0.0,1.0);
      glVertex3f(0.0,0.0,0.0);
      glTexCoord2f(0.0, 0.0);
      glVertex3f(0.0,0.0,0.0);
      glTexCoord2f(1.0,0.0);
      glVertex3f(lb.rx*gl.f,lb.ry*gl.f,lb.rz*gl.f);
      glTexCoord2f(1.0, 1.0);
      glVertex3f(r,lb.sy*gl.f,lb.sz*gl.f);
      glEnd();
      break;
  }
}

void Plot_Objects(void)
{
  int a;

  glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  if (!bg.save_new) {
    bg.new_num=body_alloc;
  }
  SetVar();
  SetTrans();
  CheckFont(0);
  for (a=0;a<bg.body_num;a++) Plot_Body(a,0);
  for (a=bg.new_num;a<body_alloc;a++) Plot_Body(a,0);
  glFlush();
  if (ll.moveit) {      /* if light is moved then plot the sun(s) */
    GLfloat suncolor[4]={1.0,1.0,0.0,1.0};

    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,suncolor);
    glEnable(GL_COLOR_MATERIAL);
    glColor4f(1.0,1.0,0.0,1.0);
    glDisable(GL_TEXTURE_2D);
    for (a=0;a<MaxLight;a++) {
      if (gl.light[a].on) {
        glLoadIdentity();
        glTranslatef(gl.x,gl.y,gl.z);
        glTranslatef(gl.light[a].param[0][0],gl.light[a].param[0][1],
                        gl.light[a].param[0][2]);
        gluSphere(gl.quad.q, 0.1, gl.quad.sl,gl.quad.st);
      }
    }
  }
  for (a=0;a<bg.body_num;a++) Plot_Body(a,1);
  for (a=bg.new_num;a<body_alloc;a++) Plot_Body(a,1);
  bg.gbeg[bg.obj_num]=bg.new_num;
  bg.gnum[bg.obj_num]=body_alloc-bg.new_num;
  CheckFont(1);
  glFlush();
  if (anim.constrreread>0) anim.constrreread=1;
}

void Print_Warning(int a,int v)
{
  static int was[2];
  int i;
#define W_AllocP 0
#define W_AllocN 1

  switch (a) {
    case 0:
      if (was[0]) break;
      fprintf(stderr,"Mozi warning: More objects (in line %d) than ",v);
      fprintf(stderr,"body_alloc(=%d). Use -b flag.\n",body_alloc);
      was[0]=1;
      break;
    case 1:
      if (was[1]) break;
      fprintf(stderr,"Mozi warning: In first line more objects (%d) than ",v);
      fprintf(stderr,"body_alloc(=%d).\nUse -b flag.\n",body_alloc);
      was[1]=1;
      break;
    case -1:
      for (i=0;i<2;i++) was[i]=0;
  }
}

void Set_FPos(void)
{
  int a;
  for (a=0;a<bg.body_num;a++) {
    body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM]=(double)animation.framepc;
    body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+2]=
		(double)animation.framepc*animation.framerepeat+
			animation.framerepeatcounter;
  }
}

void Read_Objects(int pos,int det)
{
  int a,n,d,i,ll,bs,on,obj,foursix;
  int typ,tex,num,dummyint;
  double x,y,z,rx,ry,rz,phi,time;
  char *p;

  anim.objectread=1;
  if (anim.constrreread>0) anim.constrreread=2;
  obj=0;
  ll=bg.line_length;
  if (bg.line_length<=0) {
    io.seek(io.in,animation.framepos[pos],SEEK_SET);
    if ((bg.input_type%2==0) && (bg.input_type>1)) {
      ll=animation.framepos[pos+1]-animation.framepos[pos];
      a=io.read(io.in,ll);
      if (a!=ll) {
        fprintf(stderr,"Mozi warning: in data file type %d at",bg.input_type);
        fprintf(stderr," pos %d short line (%d/%d)\n",pos,a,bg.line_length);
        return;
      }
    } else {
      /* text input */
      io.getline(io.in);
    }
    io.scan("%d",&a);time=(double)a;
  } else {
    /* we have fixed line length */
    io.seek(io.in,animation.framepos[0]+pos*bg.line_length,SEEK_SET);
    a=io.read(io.in,bg.line_length);
    if (a!=bg.line_length) {
      fprintf(stderr,"Mozi warning: in data file type %d at",bg.input_type);
      fprintf(stderr," pos %d short line (%d/%d)\n",pos,a,bg.line_length);
      return;
    }
  }
  bg.body_num=bg.wall_num;
  foursix=0;
  switch (bg.input_type) {
    case 6:
      foursix=6;
    case 4:
      i=sizeof(int);d=sizeof(double);
      p=io.buff;
      time=*((double *)p);p+=d;
      obj=bg.body_num-1;
      /* object type is byte, end if it is zero */
      while ((*p) && (p<ll+io.buff)) {
        typ=*((int *)p);p+=i;
	if (foursix) {
	  tex=*((int *)p);p+=i;
	} else  {
	  tex=*p++;
	}
        num=*((int *)p);p+=i;
        bs=bg.body_num;
        bg.body_num+=max(det,num);
        obj++;
	if (bg.gbeg!=NULL) {
	  bg.gbeg[obj]=bs;
	  bg.gnum[obj]=bg.body_num-bs;
	}
        if (bg.body_num>body_alloc) {
          Print_Warning(W_AllocP,pos);
          bg.body_num=body_alloc;
        }
        for (a=bs;a<bg.body_num;a++) {
          if (a>body_alloc) break;
	  if (foursix) {
	    body[a].tex=(tex&31);
	    body[a].color[0]=((tex>>8)&255)/255.0;
	    body[a].color[1]=((tex>>16)&255)/255.0;
	    body[a].color[2]=((tex>>24)&255)/255.0;
	    body[a].color[3]=0.5;
	  } else {
	    body[a].tex=(tex&15)+1;
	  }
          body[a].type=typ&BP_Mask;
          body[a].par=typ;
          /* determine object from different source only at reset */
          if (!anim.groupbytex) body[a].obj=obj;
          if (num==0) break;
          /* get size of body. It's compulsory */
          body[a].sx=*((double *)p);p+=d;
          if ((body[a].type==BT_Cylinder) || (body[a].type==BT_Force) ||
	      (body[a].type==BT_DTEdge) || (body[a].type==BT_Arrow)) {
            body[a].sz=*((double *)p);p+=d;
          }
          if (body[a].type==BT_BallCC) {
            body[a].params[5*CEF_SUBPARNUM]=(double)(*((int *)p));p+=i;
            body[a].sy=body[a].params[5*CEF_SUBPARNUM];
          }
          if (body[a].type==BT_Wall) {
            body[a].sy=*((double *)p);p+=d;
          }
          if ((body[a].type==BT_Force) || (body[a].type==BT_DTEdge)) {
            body[a].ind1=*((int *)p);p+=i;
            body[a].ind2=*((int *)p);p+=i;
          }
          if ((body[a].type==BT_Ellipsoid) || (body[a].type==BT_Cone) ||
	      (body[a].type==BT_Triangle) || (body[a].type==BT_Triangle2)) {
            body[a].sy=*((double *)p);p+=d;
            body[a].sz=*((double *)p);p+=d;
          }
          /* position */
          if (typ&BP_Pos) {
            body[a].x=*((double *)p);p+=d;
            body[a].y=*((double *)p);p+=d;
            body[a].z=*((double *)p);p+=d;
          } else body[a].x=body[a].y=body[a].z=0.0;
          /* orientation */
          if (typ&BP_Orient) {
            body[a].rx=*((double *)p);p+=d;
            body[a].ry=*((double *)p);p+=d;
            body[a].rz=*((double *)p);p+=d;
            body[a].phi=*((double *)p);p+=d;
          } else {
            body[a].rx=body[a].ry=body[a].phi=0.0;
            body[a].rz=1.0;
          }
          /* other parameters */
          on=5*CEF_SUBPARNUM;
          if (typ&BP_Coord) {
            bs=*((int *)p);
            body[a].params[on]=(double)bs;
            p+=i;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_Velo) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_Force) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_AngVelo) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_Torque) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_TrVelo) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_TrForce) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_TrAngVelo) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          if (typ&BP_Handle) {
            bs=*((int *)p);
            body[a].params[2]=(double)bs;
            p+=i;
	  }
          on+=CEF_SUBPARNUM;
          if (typ&BP_LifeTime) {
            body[a].params[on]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_MatPar) {
            body[a].params[on]=*((double *)p);p+=d;
            body[a].params[on+1]=*((double *)p);p+=d;
            body[a].params[on+2]=*((double *)p);p+=d;
          }
          on+=CEF_SUBPARNUM;
          if (typ&BP_TrCoord) {
            body[a].params[on]=*((double *)p);p+=d;
          }
          if (tex&16) body[a].appearance ^= BA_Texture2D;
          if (tex&32) body[a].appearance |= BA_Blend;
	  body[a].blend=(tex&32)/32;
        }
      }
      if ((det) && (!anim.groupbytex)) bg.obj_num=obj+1;
      break;
    case 3:
      perr("data file type 3 is not supported in mozi\n");
      break;
    case 2:
      perr("data file type 2 is not any more supported in mozi\n");
      i=sizeof(int);d=sizeof(double);
      p=io.buff;
      time=*((double *)p);p+=d;
      obj=bg.body_num-1;
      /* object type is byte, end if it is zero */
      while ((*p) && (p<bg.line_length+io.buff)) {
        typ=*p++;
        tex=*p++;
        num=*((int *)p);p+=i;
        obj++;
        for (a=bg.body_num;a<bg.body_num+num;a++) {
          if (a>body_alloc) break;
          body[a].type=typ;
          /* determine object from different source only at reset */
          if (!anim.groupbytex) body[a].obj=obj;
          body[a].x=*((double *)p);p+=d;
          body[a].y=*((double *)p);p+=d;
          body[a].z=*((double *)p);p+=d;
          body[a].rx=*((double *)p);p+=d;
          body[a].ry=*((double *)p);p+=d;
          body[a].rz=*((double *)p);p+=d;
          body[a].phi=*((double *)p);p+=d;
          body[a].sx=*((double *)p);p+=d;
          if (body[a].type==BT_Cylinder) {
            body[a].sz=*((double *)p);p+=d;
          }
          if (body[a].type==BT_BallCC) p+=i;
          if (body[a].type==BT_Wall) {
            body[a].sy=*((double *)p);p+=d;
          }
          body[a].tex=(tex&15)+1;
          body[a].appearance=BA_Color+BA_Texture2D;
          if (tex&16) body[a].appearance ^= BA_Texture2D;
          if (tex&32) body[a].appearance |= BA_Blend;
	  body[a].blend=(tex&32)/32;
        }
        bg.body_num+=num;
      }
      if ((det) && (!anim.groupbytex)) bg.obj_num=obj+1;
      if (bg.body_num>body_alloc) {
        Print_Warning(W_AllocP,pos);
        bg.body_num=body_alloc;
      }
      break;
    case 1:
      while ((io.scan("%d",&typ)==1)  && (bg.body_num<body_alloc)) {
        io.scan("%d",&tex); io.scan("%d",&num);
        obj--;
        for (a=bg.body_num;a<bg.body_num+num;a++) {
          if (a>body_alloc) break;
          body[a].type=typ;
          if (det) body[a].obj=obj;
          io.scan("%le",&body[a].x);
          io.scan("%le",&body[a].y);
          io.scan("%le",&body[a].z);
          io.scan("%le",&body[a].rx);
          io.scan("%le",&body[a].ry);
          io.scan("%le",&body[a].rz);
          io.scan("%le",&body[a].phi);
          io.scan("%le",&body[a].sx);
          if (body[a].type==BT_Cylinder) io.scan("%le",&body[a].sz);
          if (body[a].type==BT_BallCC) io.scan("%d",&dummyint);
          if (body[a].type==BT_Wall) io.scan("%le",&body[a].sy);
          body[a].tex=(tex&15)+1;
          body[a].appearance=BA_Color+BA_Texture2D;
          if (tex&16) body[a].appearance ^= BA_Texture2D;
          if (tex&32) body[a].appearance |= BA_Blend;
	  body[a].blend=(tex&32)/32;
        }
        bg.body_num+=num;
      }
      if (bg.body_num>body_alloc) {
        Print_Warning(W_AllocP,pos);
        bg.body_num=body_alloc;
      }
      break;
    case 0:
    default:
      io.scan("%d",&n); /* read the number of balls */
      bg.body_num=bg.wall_num+n;
      if (bg.body_num>body_alloc) {
        Print_Warning(W_AllocP,pos);
        bg.body_num=body_alloc;
      }
      for (a=bg.wall_num;a<bg.body_num;a++) {
        io.scan("%le",&x);io.scan("%le",&y);io.scan("%le",&z);
        io.scan("%le",&rx);io.scan("%le",&ry);io.scan("%le",&rz);
        io.scan("%le",&phi);
        body[a].type=BT_Ball;
        body[a].sx=0.001;
        body[a].x=x;
        body[a].y=y;
        body[a].z=z;
        body[a].rx=rx;
        body[a].ry=ry;
        body[a].rz=rz;
        body[a].phi=phi;
        if (!anim.groupbytex) body[a].obj=bg.wall_num;
      }
      if (!anim.groupbytex) bg.obj_num=bg.wall_num+1;
      break;
  }
  if (bg.obj_num) {
    bs=0;
    for (a=0;a<bg.body_num;a++) {
      /* time is set here */
      body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM]=(double)pos;
      body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+1]=time;
      body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+2]=
	(double)pos*animation.framerepeat+animation.framerepeatcounter;
      if (anim.groupbytex) {
        while (body[a].type!=bg.param[bs].type) bs++;
        body[a].obj=bs;
      }
      body[a].params[0]=(double)a;
      body[a].params[1]=(double)body[a].obj+1.0;
      on=CEF_SUBPARNUM;
      body[a].params[on]=body[a].sx;
      body[a].params[on+1]=body[a].sy;
      body[a].params[on+2]=body[a].sz;
      body[a].params[on+3]=(double)body[a].ind1;
      body[a].params[on+4]=(double)body[a].ind2;
      on+=CEF_SUBPARNUM;
      body[a].params[on]=body[a].x;
      body[a].params[on+1]=body[a].y;
      body[a].params[on+2]=body[a].z;
      on+=CEF_SUBPARNUM;
      body[a].params[on]=body[a].rx;
      body[a].params[on+1]=body[a].ry;
      body[a].params[on+2]=body[a].rz;
      body[a].params[on+3]=body[a].phi;
      on+=CEF_SUBPARNUM;
      if (bg.input_type<5) {
        body[a].color[0]=0.8f;
        body[a].color[1]=0.6f;
        body[a].color[2]=1.0f;
        body[a].color[3]=0.4f;
      }
      body[a].params[on]=body[a].color[0];
      body[a].params[on+1]=body[a].color[1];
      body[a].params[on+2]=body[a].color[2];
      body[a].params[on+3]=body[a].color[3];
    }
  }
}

void Not_Fixed_Linelength()
{
  int i,d,d3,v,llmax,typ,num,lpos,first;
  int bases[15];
  long *buff,sumbodynum;

  i=sizeof(int);d=sizeof(double);d3=3*d;
  bases[BT_EOL]=0;bases[BT_Wall]=d+d;bases[BT_Ball]=d;
  bases[BT_Cylinder]=d+d;bases[BT_Disk]=d;bases[BT_BallCC]=d+i;
  bases[BT_Force]=d+d+i+i;bases[BT_STBall]=d;bases[BT_DGBall]=d;
  bases[BT_DTEdge]=d+d+i+i;bases[BT_Ellipsoid]=d+d+d;
  bases[BT_Arrow]=d+d;bases[BT_Cone]=d+d+d;bases[BT_Triangle]=d+d+d;
  bases[BT_Triangle2]=d+d+d;
  llmax=0;lpos=io.tell(io.in);
  sumbodynum=0;first=1;
  do {
    if (llmax<io.tell(io.in)-lpos) llmax=io.tell(io.in)-lpos;
    animation.framepos[animation.maxframe++]=io.tell(io.in);
    if (animation.maxframe==animation.framepossize) {
      /* buffer is full */
      buff=animation.framepos;
      animation.framepos=(long *)
                      calloc(animation.framepossize*2,sizeof(long));
      memcpy(animation.framepos,buff,animation.framepossize*sizeof(long));
      free(buff);
      animation.framepossize*=2;
    }
    io.seek(io.in,d,SEEK_CUR);
    do {
      io.read(io.in,i);
      typ=*((int *)io.buff);
      if (typ) {
        if (bg.input_type<5) {
          io.seek(io.in,1,SEEK_CUR);
        } else {
          io.seek(io.in,i,SEEK_CUR);
        }
        io.read(io.in,i);
        num=*((int *)io.buff);
	sumbodynum+=num;
        v=bases[typ&BP_Mask];
        if (typ&BP_Pos) v+=d3;
        if (typ&BP_Orient) v+=d3+d;
        if (typ&BP_Coord) v+=i;
        if (typ&BP_Velo) v+=d3;
        if (typ&BP_Force) v+=d3;
        if (typ&BP_AngVelo) v+=d3;
        if (typ&BP_Torque) v+=d3;
        if (typ&BP_TrVelo) v+=d3;
        if (typ&BP_TrForce) v+=d3;
        if (typ&BP_TrAngVelo) v+=d3;
        if (typ&BP_Handle) v+=i;
        if (typ&BP_LifeTime) v+=d;
        if (typ&BP_MatPar) v+=d3;
        if (typ&BP_TrCoord) v+=d;
        io.seek(io.in,v*num,SEEK_CUR);
      } else if (first) {
	if (body_alloc<sumbodynum) Print_Warning(W_AllocN,sumbodynum);
	first=0;
      }
    } while ((!io.eof(io.in)) && (typ));
  } while (!io.eof(io.in));
  if (llmax<io.tell(io.in)-lpos) llmax=io.tell(io.in)-lpos;
  animation.maxframe--;
}

void reset_objects()
{
  int a,tex,typ,p,num,size_int,size_dble,goon;
  long pos,*buff;
  char ch;

  bg.line_length=-1;
  if (!io_detfiletype()) perr("input file %s does not exist\n",io.filename);
  io.in=io.open(io.filename,"rb");
  if (io.in==NULL) perr("Input file %s not found\n",io.filename);

  body=(BODY *)calloc(body_alloc+1,sizeof(BODY));
  if (!anim.donotdoublebody) {
    restb=(RestBODY *)calloc(body_alloc+1,sizeof(RestBODY));
  }
  for (a=0;a<body_alloc;a++) {
    body[a].type=BT_Ball;
    body[a].x=0.0;
    body[a].y=0.0;
    body[a].z=0.0;
    body[a].rx=0.0;
    body[a].ry=0.0;
    body[a].rz=1.0;
    body[a].phi=0.0;
    body[a].sx=0.001;
    body[a].sy=0.0;
    body[a].sz=0.0;
    body[a].appearance=BA_Color+BA_Texture2D;
    body[a].blend=0;
    body[a].tex=2;
    body[a].par=96;
  }
  bg.wall_num=0;

  /* detect frame positions */
  animation.framepossize=1000;
  animation.framepos=(long *) calloc(animation.framepossize,sizeof(long));
  animation.maxframe=0;
  animation.framepc=0;
  bg.save_new=0;
  bg.input_type=0;goon=1;
  do {
    pos=io.tell(io.in);
    io.getline(io.in);
    ch=io.buff[0];
    if ((!io.eof(io.in)) && (ch!='#')) {
      /* skip comments */
      animation.framepos[animation.maxframe++]=pos;
      if (animation.maxframe==animation.framepossize) {
        /* buffer is full */
        buff=animation.framepos;
        animation.framepos=(long *)
                        calloc(animation.framepossize*2,sizeof(long));
        memcpy(animation.framepos,buff,animation.framepossize*sizeof(long));
        free(buff);
        animation.framepossize*=2;
      }
    } else if (ch=='#') {       /* special */
      if (strlen(io.buff)==2) { /* A single # character in a line */
        if (bg.line_length>0) {
          /* We have the position information in the header so we quit */
          animation.framepos[0]=pos=io.tell(io.in);
          animation.framepossize=
                animation.maxframe=(io.tell(io.in)-pos)/bg.line_length;
        } else {
          Not_Fixed_Linelength();
        }
        goon=0;
        break;
      }
      if (strncmp(io.buff+1,"object",6)==0) {   /* indeed */
        p=1+strlen("object");
        sscanf(io.buff+p,"%d %d %d",&typ,&tex,&num);
        p+=move_bp(io.buff+p,3);
        for (a=bg.wall_num;a<bg.wall_num+num;a++) {
          body[a].type=typ;
          sscanf(io.buff+p,"%lf %lf %lf %lf %lf %lf %lf %lf",
              &body[a].x,&body[a].y,&body[a].z,&body[a].rx,
              &body[a].ry,&body[a].rz,&body[a].phi,&body[a].sx);
          p+=move_bp(io.buff+p,8);
          if ((body[a].type==BT_Cylinder) || (body[a].type==BT_BallCC)) {
            sscanf(io.buff+p,"%lf",&body[a].sz);
            p+=move_bp(io.buff+p,1);
          }
          if (body[a].type==BT_Wall) {
            sscanf(io.buff+p,"%lf",&body[a].sy);
          }
          body[a].tex=(tex&15)+1;
          body[a].appearance=BA_Color+BA_Texture2D;
          if (tex&16) body[a].appearance ^= BA_Texture2D;
          if (tex&32) body[a].appearance |= BA_Blend;
	  body[a].blend=(tex&32)/32;
        }
        bg.wall_num+=num;
      } else if (strncmp(io.buff+1,"type",4)==0) {      /* type is different */
        p=1+strlen("type");
        sscanf(io.buff+p,"%d",&bg.input_type);
        if ((bg.input_type==3) || (bg.input_type==2)) {
          perr("data file type 2, 3 is not supported in mozi\n");
        }
        if (bg.input_type==4) {
          /* read size of int/dble, line length */
          p+=move_bp(io.buff+p,1);
          sscanf(io.buff+p,"%d %d %d",&size_int,&size_dble,&bg.line_length);
          if ((size_int!=sizeof(int)) || (size_dble!=sizeof(double))) {
            perr("Size of int or double does not match!\n");
          }
          if (io.linelen<bg.line_length) {
            /* we reallocate io.buff */
            free(io.buff);
            io.linelen=bg.line_length+1;
            io.buff=(char *)calloc(io.linelen,sizeof(char));
          }
        }
      }
    }
  } while ((!io.eof(io.in)) && (goon));
  animation.maxframe--;
  bg.body_num=bg.wall_num;
  io.rewind(io.in);
  if (animation.marker2<0) {
    animation.marker2=animation.maxframe;
  }
  bg.obj_num=0;

}

void reset_BodyDisplaySettings()
{
  int a,b,typ,tex;

  anim.constrreread=0;
  bg.gbeg=NULL;
  Read_Objects(0,1);    /* read objects we have */
  if (anim.groupbytex) {
    /* group by texture and type */
    bg.obj_num=0;typ=tex=0;
    for (a=0;a<bg.body_num;a++) {
      if ((body[a].type!=typ) || (body[a].tex!=tex)) {
	bg.obj_num++;
	typ=body[a].type;tex=body[a].tex;
      }
    }
    /* allocate memory */
    bg.param=(bodydisplaystr *) calloc(bg.obj_num+1,sizeof(bodydisplaystr));
    bg.gbeg=(int *)calloc(bg.obj_num+1,sizeof(int));
    bg.gnum=(int *)calloc(bg.obj_num+1,sizeof(int));
    b=typ=tex=0;
    for (a=0;a<bg.body_num;a++) {
      if ((body[a].type!=typ) || (body[a].tex!=tex)) {
	bg.gbeg[b]=a;
	bg.gnum[b]=0;
	typ=body[a].type;tex=body[a].tex;
	if (anim.init_show) bg.param[b].show=1;
	else bg.param[b].show=0;
	bg.param[b].type=typ;
	bg.param[b].par=body[a].par;
	bg.param[b].blend=body[a].appearance&BA_Blend;
	bg.param[b].texture=((body[a].appearance&BA_Texture2D) ? tex : 0);
	bg.param[b].color[0]=0.8f;
	bg.param[b].color[1]=0.6f;
	bg.param[b].color[2]=1.0f;
	bg.param[b].color[3]=0.4f;
	b++;
      }
      bg.gnum[b-1]++;
      body[a].obj=b-1;
    }
  } else {
    bg.param=(bodydisplaystr *) calloc(bg.obj_num+1,sizeof(bodydisplaystr));
    bg.gbeg=(int *)calloc(bg.obj_num+1,sizeof(int));
    bg.gnum=(int *)calloc(bg.obj_num+1,sizeof(int));
    for (a=0,b=0;a<bg.wall_num;a++,b++) {
      bg.gbeg[b]=a;
      bg.gnum[b]=1;
      typ=body[a].type;tex=body[a].tex;
      if (anim.init_show) bg.param[b].show=1;
      else bg.param[b].show=0;
      bg.param[b].type=typ;
      bg.param[b].par=body[a].par;
      bg.param[b].blend=body[a].appearance&BA_Blend;
      bg.param[b].texture=((body[a].appearance&BA_Texture2D) ? tex : 0);
      bg.param[b].color[0]=0.8f;
      bg.param[b].color[1]=0.6f;
      bg.param[b].color[2]=1.0f;
      bg.param[b].color[3]=0.4f;
      body[a].obj=b;
    }
    for (;b<bg.obj_num;b++) {
      for (;a<bg.body_num;a++) {
	if (body[a].obj==b) break;
      }
      bg.gbeg[b]=a;
      if (b>0) bg.gnum[b-1]=bg.gbeg[b]-bg.gbeg[b-1];
      typ=body[a].type;tex=body[a].tex;
      if (anim.init_show) bg.param[b].show=1;
      else bg.param[b].show=0;
      bg.param[b].type=typ;
      bg.param[b].par=body[a].par;
      bg.param[b].blend=body[a].appearance&BA_Blend;
      bg.param[b].texture=((body[a].appearance&BA_Texture2D) ? tex : 0);
      bg.param[b].color[0]=0.8f;
      bg.param[b].color[1]=0.6f;
      bg.param[b].color[2]=1.0f;
      bg.param[b].color[3]=0.4f;
    }
    bg.param[bg.obj_num].show=1;
    bg.param[bg.obj_num].type=2;
    bg.param[bg.obj_num].par=96;
    bg.param[bg.obj_num].blend=0;
    bg.param[bg.obj_num].texture=3;
    bg.param[bg.obj_num].color[0]=0.9f;
    bg.param[bg.obj_num].color[1]=0.7f;
    bg.param[bg.obj_num].color[2]=0.2f;
    bg.param[bg.obj_num].color[3]=1.0f;
  }
  if (bg.obj_num>1) bg.gnum[bg.obj_num-1]=bg.body_num-bg.gnum[bg.obj_num-2];
  else bg.gnum[bg.obj_num-1]=bg.body_num;
  bg.new_num=body_alloc;
  bg.gbeg[bg.obj_num]=bg.new_num;
  bg.gnum[bg.obj_num]=body_alloc-bg.new_num;
  Print_Warning(-1,0);
  Read_Objects(0,0);    /* read objects we have */
}

/* EOF */
